%{
//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	SelectLexer.l:	--- lexical analysis of selector expressions ---
//
//	DESCRIPTION:
//		This is a flex/lex input file for generating a scanner for the
//		lexical analysis of boolean expressions to be supported by evselect/
//		hkselect.
//		If flex is to be used, the macro
//			FLEX_SCANNER
//		need to be defined which triggers the usage of streamlined I/O
//		routines.
//
//	CHANGE HISTORY:
//		UL, 1997-12-17: 1.0.0: - first version
//		UL, 1998-04-06: 1.1.0: - various new features added
//		UL, 1998-04-09: 1.2.0: - ifthenelse added
//		UL, 1998-05-19: 2.0.0: - various changes to bring code in line with
//								 other changes leading to selectlib v2.0
//							   - made it more C++'ish - resulting scanner
//								 needs now to be compiled with c++ compiler
//		UL, 1998-06-18: 2.1.0: - various changes to support new features
//								 available with parser v2.1.0
//								 (see SelGrammar.y)
//							   - the symbol "SELECTED" is recognized and
//								 expanded into the construct
//									SELECT_COLUMN & SELECT_MASK == 0
//								 SELECT_COLUMN & SELECT_MASK are constant
//								 strings defined in Selector.h which will
//								 be entered into the symbol table; the
//								 Evaluator needs to replace these symbols
//								 with the name of a flag column in the
//								 event list table and a valid event selection
//								 mask value; SELECTED then evaluates to
//								 'true' for all table rows which were selected
//								 in a previous run.
//		UL, 1999-01-27: 2.2.0: - minor fix to avoid warnings under gcc 2.8.1
//		UL, 1999-08-16: 2.3.0: - added hyperbolic trig functions +
//								 fmod/modf/int/ceil/floor/pow
//		UL, 1999-08-18: 2.4.0: - bug fix in modf
//		UL, 1999-11-17: 2.5.0: - changes to avoid name clashes at the linking
//								 stage when more than one lexer/parser is
//								 present:
//									o flex is invoked with -Pselect
//									  to prefix all externally visible symbols
//									  with "select"
//									o locally added lexer function are wrapped
//									  in a namespace
//		UL, 2000-06-16: 2.6.0: - real numbers given in the form:
//									'1e2' (w/o decimal point) are recognized
//								 now; they were previously treated as symbol
//								 names
//		UL, 2000-07-03: 2.7.0: - several changes related to support for
//								 string columns and NULL values
//							   - determine now the type of a column identifier
//								 in the lexer; this is necessary to be able
//								 determine the context of expression
//								 like 'COL1==COL2' (string or numeric);
//		UL, 2000-07-08: 2.8.0: - call YY_NEW_FILE in lexinit() to put the
//								 parser back into a defined state after
//								 an error condition; this fixes
//								 SPR 1663: syntax errors
//		UL, 2000-09-13: 2.9.0: - the base-pre-fix characters 'o'/'b'/'h'
//								 to identify octal/binary/hex numbers
//								 have to be in lower case (to minimize
//								 chances of name clashes with column names)
//		UL, 2001-05-14: 2.9.1: - fixed bug in real-number pattern: caused
//									E123
//								 to be recognized as a valid constant
//		UL, 2002-05-29: 2.10.0:- swapped meaning of #RAD/#DEG
//							   - better recognition of string identifiers
//								 in textual contexts
//		UL, 2002-06-03: 2.11.0:- added recognition of time and angle
//								 parameters
//
//	AUTHOR:
//		U. Lammers (UL), Dec 17 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/SelectLexer.l,v 1.7 2002/06/03 17:03:20 ulammers Exp $
//

#include <cassert>
#include <cstring>
#include <cstdlib>
#include <string>

#include "basicdefs.h"
#include "NumericFunctors.h"
#include "StringFunctors.h"
#include "Selector.h"
namespace SelectParser {
	#include "SelectParser.h"
}
#include "SelectError.h"
#include "Celestial.h"
#include "STime.h"

static const char *FIXTIME = "1997-12-31T23:59:58.816";
using namespace std;

namespace SelectParser {

#if defined(FLEX_SCANNER)			/* we use flex */
	#undef	YY_INPUT
	#define	YY_INPUT(buf,result,max_size) \
				result = SelectParser::lexiread(buf, max_size);
	static int lexiread (char *, int);
#else
	#undef	input
	#undef	unput
	#undef	output
	static int input (void), unput (int), output (int);
#endif

#define	selectwrap()	1			// avoid usage of liblex

}

%}

a		[aA]
b		[bB]
c		[cC]
d		[dD]
e		[eE]
f		[fF]
g		[gG]
h		[hH]
i		[iI]
j		[jJ]
k		[kK]
l		[lL]
m		[mM]
n		[nN]
o		[oO]
p		[pP]
q		[qQ]
r		[rR]
s		[sS]
t		[tT]
u		[uU]
v		[vV]
w		[wW]
x		[xX]
y		[yY]

white	[ \t\n]
alpha	[a-zA-Z_]
alnum	[a-zA-Z_0-9]
/* alphanumeric + dot (.) and dash (-) */
alnumx	[a-zA-Z_0-9\.\-]

	/* ------------( binary, octal, decimal, hexadecimal digits )---------*/
bdigit	[01]
odigit	[0-7]
ddigit	[0-9]
hdigit	[0-9aAbBcCdDeEfF]

	/* ---------(binary, octal, decimal, hexadecimal, real numbers )------*/
binnum	(b{bdigit}+)
octnum	(o{odigit}+)
decnum	({ddigit}+)
hexnum	((h|(0{x})){hdigit}+)
exp		({e}[+-]?{ddigit}+)
realnum	((({ddigit}+\.{ddigit}*)|({ddigit}*\.?{ddigit}+)){exp}?)
ident	({alpha}{alnum}*)
msnum	([0-5]?{ddigit})
optfrac	((\.{ddigit}+)?)

	/* -----------------( string literals )-------------------------------*/
	/* regular expression matching C-strings, see "Mastering Regular	  */
	/* Expression", O'Reilly, p. 168									  */
stringliteral	((\"[^"\\]*(\\.[^"\\]*)*\")|(\'[^\']*\'))

	/* --------------------( angles )-------------------------------------*/
minsec	({msnum}{m}({msnum}({optfrac}s))?)

ra		({decnum}{h}{minsec}?)
dec		(-?{realnum}{d}{minsec}?)

	/* ------------------------( times )---------------------------------*/
yearnum	([0-9]{1,4})
monnum	((0?[1-9])|(1[0-2]))
daynum	((0?[1-9])|([12][0-9])|3[01])
hournum	((0[0-9])|(1[0-9])|(2[0-3]))
weekday	({m}{o}{n}|{t}{u}{e}|{w}{e}{d}|{t}{h}{u}|{f}{r}{i}|{s}{a}{t}|{s}{u}{n})
monname	({j}{a}{n}|{f}{e}{b}|{m}{a}{r}|{a}{p}{r}|{m}{a}{y}|{j}{u}{n}|{j}{u}{l}|{a}{u}{g}|{s}{e}{p}|{o}{c}{t}|{n}{o}{v}|{d}{e}{c})

fitstime	({yearnum}-{monnum}-{daynum}(T{hournum}(:{msnum}(:{msnum}{optfrac})?)?)?)
jdnum	({j}{d}{realnum})
mjdnum	({m}{jdnum})
dynum	({y}{realnum})
caldate	({weekday}{white}+{monname}{white}+{daynum}{white}+{hournum}:{msnum}:{msnum}{white}+{yearnum})

	/* ----------( true, false, ifthenelse, selected )--------------------*/
true	({t}{r}{u}{e})
false	({f}{a}{l}{s}{e})
ifthenelse	({i}{f}{t}{h}{e}{n}{e}{l}{s}{e})
selected	({s}{e}{l}{e}{c}{t}{e}{d})

	/* -----------------------( isnull )-------------------------*/
isnull	({i}{s}{n}{u}{l}{l})

	/* -----------------( arithmetic functions )-------------------------*/
fabs	({a}{b}{s})

fsin	({s}{i}{n})
fcos	({c}{o}{s})
ftan	({t}{a}{n})
arc		({a}{r}{c})
asin	({arc}{fsin})
acos	({arc}{fcos})
atan	({arc}{ftan})
atan2	({arc}{ftan}"2")

cosh	({fcos}{h})
sinh	({fsin}{h})
tanh	({ftan}{h})

fexp	({e}{x}{p})
flog	(({l}{o}{g})|({l}{n}))
flog10	({flog}"10")
fsqrt	({s}{q}{r}{t})

modf	({m}{o}{d}{f})
pow		({p}{o}{w})
ceil	({c}{e}{i}{l})
floor	({f}{l}{o}{o}{r})
fmod	({f}{m}{o}{d})
fint	({i}{n}{t})

point	({p}{o}{i}{n}{t})
line	({l}{i}{n}{e})
box		({b}{o}{x})
rect	({r}{e}{c}{t}{a}{n}{g}{l}{e})
diamond	(({d}{i}{a}{m}{o}{n}{d})|({r}{h}{o}{m}{b}{u}{s}))
sector	(({p}{i}{e})|({s}{e}{c}{t}{o}{r}))
circle	({c}{i}{r}{c}{l}{e})
ring	(({r}{i}{n}{g})|({a}{n}{n}{u}{l}{u}{s}))
ellip	({e}{l}{l}{i}{p})
ellipse	({ellip}{s}{e})
elliptr	({ellip}{t}{ring})
polygon	({p}{o}{l}{y}{g}{o}{n})
polygon2 ({polygon}"2")
region	({r}{e}{g}{i}{o}{n})
near	({n}{e}{a}{r})
vector	({v}{e}{c}{t}{o}{r})
skyvec	({s}{k}{y}{vector})
unitvec	({u}{n}{i}{t}{vector})
cone	({c}{o}{n}{e})
crossprod ({c}{r}{o}{s}{s})
norm	({n}{o}{r}{m})

	/* -----------------( string functions )-------------------------*/
lower	({l}{o}{w}{e}{r})
upper	({u}{p}{p}{e}{r})
strlen	({s}{t}{r}{l}{e}{n})
ascii	({a}{s}{c}{i}{i})

/* optional whitespace followed by '(' */
wpar	({white}*"(")
wbar	({white}*"[")

in		({i}{n})
gti		({g}{t}{i})
mask	({m}{a}{s}{k})
path	((\/?{alnumx}*)*\/?)
/* extensiom number */
extspec	((\+{alnumx}+)|(\[{alnumx}+\])|(:{alnumx}+))
/* file name */
fname	({path}{alnumx}*{extspec}?)
/* attribute (=header keyword) name */
attname	([a-zA-Z][0-9a-zA-Z_\-]*)

%s EXPECTFNAME
%s EXPECTSTRING

%%
{white}{in}{white}		return IN;				/* IN operator				*/
{white}+				;						/* skip whitespace			*/

	/* ------------( arithmetic operators )------------------------- */
"+"						return PLUS;			/* arith. addition			*/
"-"						return MINUS;			/* arith. subtraction		*/
"*"						return MULT;			/* arith. multiplication	*/
"/"						return DIV;				/* arith. division			*/
"**"					return EXPO;			/* arith. exponentation		*/
"%"						return MODULO;			/* arith. modulo			*/

	/* --------------------( logical operators )---------------------*/
"&&"|"."{a}{n}{d}"."	return ANDIF;			/* logical and				*/
"||"|"."{o}{r}"."		return ORIF;			/* logical or				*/
"!"|"."{n}{o}{t}"."		return NOT;				/* logical not				*/


	/* --------------------( bitwise operators )---------------------*/
"&"						return AND;				/* bitwise and				*/
"|"						return OR;				/* bitwise or				*/
"^"						return XOR;				/* exclusive or				*/
"<<"					return LSHIFT;			/* bitshift left			*/
">>"					return RSHIFT;			/* bitshift right			*/
"~"						return BITCOMPL;		/* one's complement			*/

	/* ------------------( relational operators )--------------------*/
"<"|"."{l}{t}"."		{ return LT;}
"<="|"."{l}{e}"."		{ return LE;}
">"|"."{g}{t}"."		{ return GT;}
">="|"."{g}{e}"."		{ return GE;}
"=="|"."{e}{q}"."		{ return EQ; }
"!="|"."{n}{e}"."		{ return NE; }

	/* ------------------( arithmetic functions )---------------------*/
{fabs}/{wpar}			{ SelectParser::yylval.yNode = new Fabs();
						  return AFUNC; }
{fsin}/{wpar}			{ SelectParser::yylval.yNode = new Fsin();
						  return AFUNC; }
{fcos}/{wpar}			{ SelectParser::yylval.yNode = new Fcos();
						  return AFUNC; }
{ftan}/{wpar}			{ SelectParser::yylval.yNode = new Ftan();
						  return AFUNC; }
{asin}/{wpar}			{ SelectParser::yylval.yNode = new Arcsin();
						  return AFUNC; }
{acos}/{wpar}			{ SelectParser::yylval.yNode = new Arccos();
						  return AFUNC; }
{atan2}/{wpar}			{ SelectParser::yylval.yNode = new Arctan2();
						  return AFUNC2; }
{atan}/{wpar}			{ SelectParser::yylval.yNode = new Arctan();
						  return AFUNC; }

{sinh}/{wpar}			{ SelectParser::yylval.yNode = new Fsinh();
						  return AFUNC; }
{cosh}/{wpar}			{ SelectParser::yylval.yNode = new Fcosh();
						  return AFUNC; }
{tanh}/{wpar}			{ SelectParser::yylval.yNode = new Ftanh();
						  return AFUNC; }

{fexp}/{wpar}			{ SelectParser::yylval.yNode = new Fexp();
						  return AFUNC; }
{flog}/{wpar}			{ SelectParser::yylval.yNode = new Flog();
						  return AFUNC; }
{flog10}/{wpar}			{ SelectParser::yylval.yNode = new Flog10();
						  return AFUNC; }
{fsqrt}/{wpar}			{ SelectParser::yylval.yNode = new Fsqrt();
						  return AFUNC; }

{modf}/{wpar}			{ SelectParser::yylval.yNode = new Modf();
						  return AFUNC; }
{pow}/{wpar}			{ SelectParser::yylval.yNode = new Pow();
						  return AFUNC2; }
{ceil}/{wpar}			{ SelectParser::yylval.yNode = new Ceil();
						  return AFUNC; }
{floor}/{wpar}			{ SelectParser::yylval.yNode = new Floor();
						  return AFUNC; }
{fmod}/{wpar}			{ SelectParser::yylval.yNode = new Fmod();
						  return AFUNC2; }
{fint}/{wpar}			{ SelectParser::yylval.yNode = new Fint();
						  return AFUNC; }

	/* ------------------( vector functions )---------------------*/
{vector}/{wpar}			{ return VECTOR; }
{skyvec}/{wpar}			{ return SKY; }
{unitvec}/{wpar}		{ return UNIT; }
{cone}/{wpar}			{ return CONE; }
{crossprod}/{wpar}		{ return CROSSPROD; }
{norm}/{wpar}			{ return NORM; }

	/* -----------------------( isnull )--------------------------*/
{isnull}/{wpar}			{ return ISNULL; }

	/* ------------------( string functions )---------------------*/
{upper}/{wpar}		    { SelectParser::yylval.yNode = new Upper();
						  BEGIN EXPECTSTRING; return SFUNC; }
{lower}/{wpar}			{ SelectParser::yylval.yNode = new Lower();
						  BEGIN EXPECTSTRING; return SFUNC; }
{strlen}/{wpar}			{ SelectParser::yylval.yNode = new Strlen();
						  BEGIN EXPECTSTRING; return ASFUNC; }
{ascii}/{wpar}			{ SelectParser::yylval.yNode = new Ascii();
						  BEGIN EXPECTSTRING; return ASFUNC; }

	/* ---------------( near, ifthenelse, selected )----------------------*/
{near}/{wpar}			{ SelectParser::yylval.yNode = new Fnear();
						  return NEAR;	}
{ifthenelse}/{wpar}		{ SelectParser::yylval.yNode = new Ifthenelse();
						  return IFTHENELSE; }
{selected}({white}*|"()") {
						  // SELECTED token will be converted into expression
						  //		flagColumn & mask != 0
						  SelectParser::yylval.yNode = new Ne(new And(
							  selectorGlobals->symbolTable->
											identifier(SELECT_COLUMN),
							  selectorGlobals->symbolTable->
											identifier(SELECT_MASK)),
							  new NumericLiteral(0.));
						  return SELECTED; }

	/* -------------------( region shape functions )----------------------*/
{point}/{wpar}			{ SelectParser::yylval.yNode = new Fpoint();
						  return SPATFUNC4; }
{line}/{wpar}			{ SelectParser::yylval.yNode = new Fline();
						  return SPATFUNC6; }
{circle}/{wpar}			{ SelectParser::yylval.yNode = new Fcircle();
						  return SPATFUNC5; }
{ring}/{wpar}			{ SelectParser::yylval.yNode = new Fring();
						  return SPATFUNC6; }
{ellipse}/{wpar}		{ SelectParser::yylval.yNode = new Fellipse();
						  return SPATFUNC7; }
{elliptr}/{wpar}		{ SelectParser::yylval.yNode = new Frellipse();
						  return SPATFUNC10; }
{box}/{wpar}			{ SelectParser::yylval.yNode = new Fbox();
						  return SPATFUNC7; }
{rect}/{wpar}			{ SelectParser::yylval.yNode = new Frectangle();
						  return SPATFUNC7; }
{polygon}/{wpar}		{ return POLYGONFUNC; }
{polygon2}/{wpar}		{ return POLYGONFUNC2; }
{sector}/{wpar}			{ SelectParser::yylval.yNode = new Fsector();
						  return SPATFUNC6; }
{diamond}/{wpar}		{ SelectParser::yylval.yNode = new Fdiamond();
						  return SPATFUNC7; }
	/* ---------------------(vector functions )---------------------------*/


	/* ---------------------(block filters )------------------------------*/
{gti}/"("				{ BEGIN EXPECTFNAME; return GTI; }
{mask}/"("				{ BEGIN EXPECTFNAME; return MASK; }
{region}/"("			{ BEGIN EXPECTFNAME; return REGION; }
<EXPECTFNAME>{fname}	{ BEGIN INITIAL; SelectParser::yylval.yFileName = new
								std::string(yytext); return FILENAME; }

	/* -----------------( idenifier and constants )----------------------*/
{binnum}				{ SelectParser::yylval.yNode = new NumericLiteral(DBL(
							(std::strtoul(yytext+1, /* skip the 'b' */
							 (char **)NULL, 2)))); return NLITERAL; }

{octnum}				{ SelectParser::yylval.yNode = new NumericLiteral(DBL(
							(std::strtoul(yytext+1,/* skip the preceding '0' */
							 (char **)NULL, 8)))); return NLITERAL; }

{decnum}				{ SelectParser::yylval.yNode = new NumericLiteral(DBL(
							(strtoul(yytext, (char **)NULL, 10))));
						  return NLITERAL; }

{hexnum}				{ SelectParser::yylval.yNode =
													new NumericLiteral((double)
							(strtoul(yytext+(toupper(yytext[0])=='H' ? 1 : 2),
							  /* skip the preceing '0x|h' */
							 (char **)NULL, 16)));
						  return NLITERAL; }

{realnum}				{ SelectParser::yylval.yNode =
											new NumericLiteral(atof(yytext));
						  return NLITERAL; }

{ra}					{ RightAscension ra(yytext);
						  SelectParser::yylval.yNode =
											new NumericLiteral(ra.angle());
						  return NLITERAL;
						}

{dec}					{ Declination dec(yytext);
						  SelectParser::yylval.yNode =
											new NumericLiteral(dec.angle());
						  return NLITERAL;
						}

{fitstime}				|
{caldate}				{ STime t(yytext);
						  SelectParser::yylval.yNode =
   										new NumericLiteral(t-STime(FIXTIME));
						  return NLITERAL;
						}

{jdnum}					{ STime t(atof(yytext+2), STime::JD);
						  SelectParser::yylval.yNode =
   										new NumericLiteral(t-STime(FIXTIME));
						  return NLITERAL;
						}

{mjdnum}				{ STime t(atof(yytext+3), STime::MJD);
						  SelectParser::yylval.yNode =
   										new NumericLiteral(t-STime(FIXTIME));
						  return NLITERAL;
						}

{dynum}					{ STime t(atof(yytext+1), STime::DYN);
						  SelectParser::yylval.yNode =
   										new NumericLiteral(t-STime(FIXTIME));
						  return NLITERAL;
						}

"#"{p}{i}				{ SelectParser::yylval.yNode =
												new NumericLiteral(M_PI);
						  return NLITERAL; }
"#"{e}					{ SelectParser::yylval.yNode = new NumericLiteral(M_E);
						  return NLITERAL; }
"#"{d}{e}{g}			{ SelectParser::yylval.yNode =
												new NumericLiteral(RAD_TO_DEG);
															return NLITERAL; }
"#"{a}{r}{c}{s}{e}{c}	{ SelectParser::yylval.yNode =
										new NumericLiteral(DEG_TO_RAD/3600.);
															return NLITERAL; }
"#"{a}{r}{c}{m}{i}{n}	{ SelectParser::yylval.yNode =
										new NumericLiteral(DEG_TO_RAD/60.);
															return NLITERAL; }
"#"{r}{a}{d}			{ SelectParser::yylval.yNode =
									new NumericLiteral(DEG_TO_RAD);
						  return NLITERAL; }
"#"({r}{o}{w})?			{ SelectParser::yylval.yNode =
							selectorGlobals->symbolTable->
								identifier(ROWNUMBER_ID);
						  return RIDENT; }
												/* the current row number  */

{true}					{ return TRUECONST; }
{false}					{ return FALSECONST; }

{stringliteral}			{ yytext[strlen(yytext)-1] = '\0';
						  SelectParser::yylval.yNode =
											new StringLiteral(yytext+1);
						  BEGIN INITIAL;
						  return SLITERAL; }


<EXPECTSTRING>{ident}	|
{ident}/{wbar}			{ string idName(yytext);
						  SelectParser::yylval.yNode =
							selectorGlobals->symbolTable->
									identifier(idName);
							
						  BEGIN INITIAL;
						  const TableDescriptor *tab =
											selectorGlobals->tableDescriptor;
						  if (tab)
							  if (tab->hasColumn(idName)) {
								  ColumnDescriptor *col = tab->column(idName);
								  if (!col->isNumeric()) {
									  SelectParser::yylval.yNode->
										  textLengthMax(col->length());
									  return SIDENT;
								  }
							  } else {
								  using namespace SelectError;
								  selerrstr<<"No column named `"
										   <<idName<<"' in table `"
										   <<tab->setName()<<":"<<tab->name()
										   <<"'\n";
							  }

						  SelectParser::yylval.yNode->textLengthMax(128);
						  return SIDENT; }

{ident}					{ string idName(yytext);
						  SelectParser::yylval.yNode =
							selectorGlobals->symbolTable->
									identifier(idName);
						  const TableDescriptor *tab =
											selectorGlobals->tableDescriptor;
						  if (tab)
							  if (tab->hasColumn(idName)) {
								  ColumnDescriptor *col = tab->column(idName);
								  if (col->isBool())
									  return BIDENT;
								  else if (col->isInteger())
									  return IIDENT;
								  else if (col->isReal())
									  return RIDENT;
								  else {
									  /* this is a string column */
									  SelectParser::yylval.yNode->
										  textLengthMax(col->length());
									  return SIDENT;
								  }
							  } else {
								  using namespace SelectError;
								  selerrstr<<"No column named `"
										   <<idName<<"' in table `"
										   <<tab->setName()<<":"<<tab->name()
										   <<"'\n";
							  }

							/* no table has been registered - assume the
							   column is real */
							return RIDENT;
						}

.						{ return yytext[0]; }

%%

/*
 *	line I/O for the scanner
 */
static char *buf, *bp;

namespace SelectParser {

void lexinit(const std::string &expression)
{
	buf = bp = const_cast<char *>(expression.c_str());
#ifdef	FLEX_SCANNER
//	yy_set_bol(1)
	YY_NEW_FILE;		//	necessary to recover from previous error
	BEGIN(INITIAL);
#endif
}

#ifdef	FLEX_SCANNER

static int lexiread (char *line, int max)
{	int n;

	for (n=0; *bp && n<max; ++n)
		*line++ = *bp++;

	return n;
}

//
//	declaration of dum1/dum2 is workaround to avoid compiler warning
//	about yy_flex_realloc/yyunput being defined but not used
//	should be replaced by compiler pragma to turn that message of
//
void *(*dum1)(void *, unsigned) = yy_flex_realloc;
void (*dum2)(int, char *) = yyunput;

#else

static int input (void)
{
	return * bp ? * bp ++ : * bp;
}

static int unput (int ch)
{
	return ch ? (*--bp = ch) : (*bp = ch);
}

static int output (int ch)
{
	assert(0);
}

#endif	/* FLEX_SCANNER */

void yyerror(char *s)
{
	using namespace SelectError;

	selerrstr<<buf<<endl;
	for (char *cp = buf; cp < bp; ++cp)
		selerrstr<<(*cp == '\t' ? '\t' : ' ');
	selerrstr<<"^ "<<s<<endl;
}

}
