%{
//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	SelectParser.y --- grammar for selection library
//
//	DESCRIPTION:
//		This is a yacc/bison input file which defines the grammar of
//		boolean expressions to be used by the event selection tasks in the SAS
//
//	CHANGE HISTORY:
//		UL, 1997-12-17: 1.0.0: - first version
//		UL, 1998-04-06: 1.1.0: - various new features addded
//		UL, 1998-04-09: 1.2.0: - ifthenelse added
//		UL, 1998-05-19: 2.0.0: - made it more C++'ish
//							   - have no a start symbol which reduces to
//								 either a logic or arithmetic expression
//		UL, 1998-06-18: 2.1.0: - added 'IN' operator for
//									o intervals, e.g. "PHA in 0.1,10"
//									o regions, e.g. (10,10) in circle(10,10,3)
//									o file-based filters, e.g.,
//										"10.1 in gti(gti.fits)"
//							   - for the file-based filters, blocks can now
//								 be specified by name as well as number, e.g.
//									gti.fits[GTI10]  or   gti.fits[12]
//							   - new boolean function "SELECTED"; gets
//								 expanded into conventional constructs in the
//								 lexer
//		UL, 1999-11-17: 2.2.0: - updates to avoid name clashes at the linking
//								 stage with more than one lexer/parser in the
//								 system:
//									o call the lexer with a special name
//									o wrap the parser in a namespace
//		UL, 2000-05-29: 2.3.0: - fixed bug in mask filter which caused
//									mask(m.ds, xOff, yOff, xCol, yCol)
//								 and
//									(xCol, yCol) in mask(m.ds, xOff, yOff)
//								 to behave differently
//		UL, 2000-07-03: 2.4.0: - several changes related to support of
//								 string columns / null values
//
//	AUTHOR:
//		U. Lammers (UL), Dec 17 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/SelectParser.y,v 1.8 2002/05/29 13:29:18 ulammers Exp $
///

#include <string>

#include "basicdefs.h"
#include "Selector.h"
#include "NumericFunctors.h"
#include "VectorFunctors.h"
#include "RegionFilter.h"
#include "StringFunctors.h"
#include "SelectBlockFuncs.h"

SelectorGlobals selectorGlobalsInstance = {
		0,		//	pointer to symbol table
		0,		//	expression tree
		true,	//	tree can be evaluated in logical context
		0,		//	table descriptor
		0		//	evaluator
};
SelectorGlobals *selectorGlobals = &selectorGlobalsInstance;

static void prependSetName(std::string *s)
{
	if ((*s)[0] == '[' || (*s)[0] == '+')
		s->insert(0, selectorGlobals->tableDescriptor ?
						selectorGlobals->tableDescriptor->setName() : "");
}

#define yylex		selectlex
extern int selectlex();			//	this is the name of the lexer function
								//	that flex generates with the -Pselect
								//	option; we do need a distinct name
								//	in order to avoid name clashes with
								//	other lexers in the SAS

namespace SelectParser {

extern void yyerror(char *);	//	The locally added functions in the
								//	lexer are in a namespace SelectParser

%}

%union {
	OpNode		yNode;
	std::string	*yFileName;
} // two different return types

	/* ----------------------( tokens )----------------------------------*/
%token PLUS, MINUS, MULT, DIV, EXPO, MODULO
%token ANDIF, ORIF, NOT, GTI, MASK, REGION, IN
%token AND, OR, XOR, LSHIFT, RSHIFT, BITCOMPL, TRUECONST, FALSECONST
%token POLYGONFUNC, POLYGONFUNC2
%token VECTOR, SKY, UNIT, CROSSPROD, CONE, NORM
%token <yNode>		LT, LE, GT, GE, EQ, NE
%token <yNode>		AFUNC, AFUNC2
%token <yNode>		SFUNC, ASFUNC
%token <yNode>		SPATFUNC4, SPATFUNC5, SPATFUNC6, SPATFUNC7, SPATFUNC10
%token <yNode>		NEAR, IFTHENELSE
%token <yNode>		BIDENT, IIDENT, RIDENT, SIDENT
%token <yNode>		NLITERAL, SLITERAL
%token <yNode>		ISNULL
%token <yNode>		SELECTED
%token <yFileName>	FILENAME
%type  <yNode>		arith, logic, vector, pointlist, intervallist, intervalspec
%type  <yNode>		text, textrange

   /* --( logical/arithmetic operators; incresing precedence )--*/
%left	ORIF					/* ||							*/
%left	ANDIF					/* &&							*/
%left	OR						/* |							*/
%left	XOR						/* ^							*/
%left	AND						/* &							*/
%left	LT LE GT GE EQ NE		/* comparisons					*/
%left	LSHIFT RSHIFT			/* <<, >>						*/
%left	PLUS MINUS				/* plus, minus					*/
%left	MULT DIV MODULO			/* multiplication, division		*/
%left	EXPO					/* exponentiation				*/
%right	NOT BITCOMPL			/* sign, negation, etc.			*/

%%

expr	: logic					{ selectorGlobals->selExpression = $1;
								  selectorGlobals->isSelExprLogic = true; }
		| arith					{ selectorGlobals->selExpression = $1;
								  selectorGlobals->isSelExprLogic = false; }

logic	: TRUECONST				{ $$ = new TrueNode; }
		| FALSECONST			{ $$ = new FalseNode; }
		| SELECTED				{ $$ = $1; }
		| logic ORIF logic		{ $$ = new OrIf($1, $3); }
		| logic ANDIF logic		{ $$ = new AndIf($1, $3); }
		| NOT logic				{ $$ = new Not($2); }
		| arith EQ arith		{ $$ = new Eq($1, $3); }
		| arith EQ logic		{ $$ = new Eq($1, $3); }
		| logic EQ arith		{ $$ = new Eq($1, $3); }
		| logic EQ logic		{ $$ = new Eq($1, $3); }
		| arith NE arith		{ $$ = new Ne($1, $3); }
		| arith NE logic		{ $$ = new Ne($1, $3); }
		| logic NE arith		{ $$ = new Ne($1, $3); }
		| logic NE logic		{ $$ = new Ne($1, $3); }
		| arith LT arith		{ $$ = new Lt($1, $3); }
		| arith LE arith		{ $$ = new Le($1, $3); }
		| arith GT arith		{ $$ = new Gt($1, $3); }
		| arith GE arith		{ $$ = new Ge($1, $3); }
		| '(' logic ')'			{ $$ = $2; }
		| text EQ text			{ $$ = new StrEq($1, $3); }
		| text NE text	        { $$ = new StrNe($1, $3); }
		| text LT text			{ $$ = new StrLt($1, $3); }
		| text LE text			{ $$ = new StrLe($1, $3); }
		| text GT text			{ $$ = new StrGt($1, $3); }
		| text GE text			{ $$ = new StrGe($1, $3); }
		| NEAR '(' arith ',' arith ',' arith ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| IFTHENELSE '(' logic ',' logic ',' logic ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| SPATFUNC4 '(' arith ',' arith ',' arith ',' arith ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7, $9 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| '(' arith ',' arith ')' IN SPATFUNC4 '(' arith ',' arith ')'
								{ $$ = $7;
								  OpNode nd[] = { $9, $11, $2, $4 };
								  $7->addChildren(NUM_OF(nd), nd); }
		| SPATFUNC5 '(' arith ',' arith ',' arith ',' arith ',' arith ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7, $9, $11 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| '(' arith ',' arith ')' IN SPATFUNC5 '(' arith ',' arith ','
										arith ')'
								{ $$ = $7;
								  OpNode nd[] = { $9, $11, $13, $2, $4 };
								  $7->addChildren(NUM_OF(nd), nd); }
		| SPATFUNC6 '(' arith ',' arith ',' arith ',' arith ','
				   	arith ',' arith ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7, $9, $11, $13 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| '(' arith ',' arith ')' IN SPATFUNC6 '(' arith ',' arith ','
										arith ',' arith ')'
								{ $$ = $7;
								  OpNode nd[] = { $9, $11, $13, $15, $2, $4 };
								  $7->addChildren(NUM_OF(nd), nd); }
		| SPATFUNC7 '(' arith ',' arith ',' arith ',' arith ',' arith ','
			arith ',' arith ')'	{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7, $9, $11, $13,
								  $15 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| '(' arith ',' arith ')' IN SPATFUNC7 '(' arith ',' arith ','
										arith ',' arith ',' arith ')'
								{ $$ = $7;
								  OpNode nd[] = { $9, $11, $13, $15, $17,
												  $2, $4 };
								  $7->addChildren(NUM_OF(nd), nd); }
		| SPATFUNC10 '(' arith ',' arith ',' arith ',' arith ',' arith ','
			arith ',' arith ',' arith ',' arith ',' arith ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5, $7, $9, $11, $13,
												  $15, $17, $19, $21 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| '(' arith ',' arith ')' IN SPATFUNC10 '(' arith ',' arith ','
										arith ',' arith ',' arith ','
										arith ',' arith ',' arith ')'
								{ $$ = $7;
								  OpNode nd[] = { $9, $11, $13, $15, $17, $19,
												  $21, $23, $2, $4 };
								  $7->addChildren(NUM_OF(nd), nd); }
		| POLYGONFUNC '(' pointlist ',' arith ',' arith ')'
								{ $$ =
									PolygonScope::newPolyFilter($3, $5, $7); }
		| '(' arith ',' arith ')' IN POLYGONFUNC '(' pointlist ')'
					   		    { $$ =
									PolygonScope::newPolyFilter($9, $2, $4); }
		| POLYGONFUNC2 '(' pointlist ',' arith ',' arith ')'
								{ $$ =
									PolygonScope::newPolyFilter($3, $5, $7,
																true); }
		| '(' arith ',' arith ')' IN POLYGONFUNC2 '(' pointlist ')'
					   		    { $$ =
									PolygonScope::newPolyFilter($9, $2, $4,
																true); }
		| GTI '(' FILENAME ',' arith ')'
								{ prependSetName($3);
								  $$ = GtiFilter::expandGtis(*$3, $5);
								  delete $3; }
		| arith IN GTI '(' FILENAME ')'
								{ prependSetName($5);
								  $$ = GtiFilter::expandGtis(*$5, $1);
								  delete $5; }
		| REGION '(' FILENAME ',' arith ',' arith ')'
								{ prependSetName($3);
								  $$ = RegionFilter::expandRegion(*$3, $5, $7);
								  delete $3; }
		| REGION '(' FILENAME ')'
								{ prependSetName($3);
								  $$ = RegionFilter::expandRegion(*$3);
								  delete $3; }
		| '(' arith ',' arith ')' IN REGION '(' FILENAME ')'
								{ prependSetName($9);
								  $$ = RegionFilter::expandRegion(*$9, $2, $4);
								  delete $9; }
		| MASK '(' FILENAME ',' arith ',' arith ',' arith ',' arith ')'
								{ prependSetName($3);
								  $$ = new MaskFilter(*$3, $5, $7, $9, $11);
								  delete $3; }
		| '(' arith ',' arith ')' IN MASK '(' FILENAME ',' arith ',' arith ')'
								{ prependSetName($9);
								  $$ = new MaskFilter(*$9, $11, $13, $2, $4);
								  delete $9; }
		| arith IN intervallist	{ $$ = $3;
								  OpNode nd[] = { $1 };
								  $3->addChildren(NUM_OF(nd), nd); }
		| ISNULL '(' IIDENT ')' { const TableDescriptor *tab =
									selectorGlobals->tableDescriptor;
								  if (tab &&
									  tab->column($3->name())->nullDefined())
									  $$ = new Eq($3, new NumericLiteral(DBL(
										  tab->column($3->name())->
															nullValue())));
								  else
									  $$ = new FalseNode; }
		| ISNULL '(' RIDENT ')' { $$ = new NanOp($3); }
		| BIDENT
		| arith EQ text			{ $1->textLengthMax(32); $$ = new FalseNode; }
		| text EQ arith			{ $3->textLengthMax(32); $$ = new FalseNode; }
		| arith NE text			{ $1->textLengthMax(32); $$ = new FalseNode; }
		| text NE arith			{ $3->textLengthMax(32); $$ = new FalseNode; }
		| CONE '(' vector ',' arith ',' vector ')'
								{ $$ = new ConeFilter($3, $5, $7); }
		| vector IN CONE '(' vector ',' arith ')'
								{ $$ = new ConeFilter($5, $7, $1); }

arith	: arith PLUS arith		{ $$ = new Add($1, $3); }
		| arith MINUS arith		{ $$ = new Sub($1, $3); }
		| arith MULT arith		{ $$ = new Mul($1, $3); }
		| arith DIV arith		{ $$ = new Div($1, $3); }
		| arith MODULO arith	{ $$ = new Modulo($1, $3); }
		| arith EXPO arith		{ $$ = new Expo($1, $3); }
		| arith AND arith		{ $$ = new And($1, $3); }
		| arith OR arith		{ $$ = new Or($1, $3); }
		| arith XOR arith		{ $$ = new Xor($1, $3); }
		| arith LSHIFT arith	{ $$ = new Lshift($1, $3); }
		| arith RSHIFT arith	{ $$ = new Rshift($1, $3); }
		| BITCOMPL arith		{ $$ = new Bitcompl($2); }
		| AFUNC '(' arith ')'	{ $$ = $1;
								  OpNode nd[] = { $3 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| AFUNC2 '(' arith ',' arith ')'
								{ $$ = $1;
								  OpNode nd[] = { $3, $5 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| MINUS arith %prec NOT	{ $$ = new Uminus($2); }
		| ASFUNC '(' text ')' { $$ = $1;
								  OpNode nd[] = { $3 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| '(' arith ')'			{ $$ = $2; }
		| vector '[' NLITERAL ']' { $$ = new VectorComponent($1, $3); }
		| vector MULT vector	{ $$ = new ScalarProduct($1, $3); }
		| NORM '(' vector ')'	{ $$ = new VectorNorm($3); }
		| NLITERAL
		| IIDENT
		| RIDENT

vector	: vector PLUS vector		{ $$ = new VectorAdd($1, $3); }
		| vector MINUS vector		{ $$ = new VectorSub($1, $3); }
		| arith MULT vector			{ $$ = new VectorScalarMult($3, $1); }
		| vector MULT arith			{ $$ = new VectorScalarMult($1, $3); }
		| vector DIV arith			{ $$ = new VectorScalarDiv($1, $3); }
		| CROSSPROD '(' vector ',' vector ')'
									{ $$ = new CrossProd($3, $5); }
		| MINUS vector %prec NOT	{ $$ = new VectorMinus($2); }
		| UNIT '(' arith ',' arith ',' arith ')'
									{ $$ = new UnitVector($3, $5, $7); }
		| SKY '(' arith ',' arith ')'
									{ $$ = new SkyVector($3, $5); }
		| VECTOR '(' arith ',' arith ',' arith ')'
									{ $$ = new Vector($3, $5, $7); }

textrange: '[' ':' ']'			{ $$ = new StrRange(false, false); }
		| '[' arith ':' ']'		{ $$ = new StrRange(true, false, $2); }
		| '[' ':' arith ']'		{ $$ = new StrRange(false, true, $3); } 
		| '[' arith ':' arith ']' { $$ = new StrRange(true, true, $2, $4); }

text	: SFUNC '(' text ')'	{ $$ = $1;
								  $1->textLengthMax($3->textLength());
								  OpNode nd[] = { $3 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| text textrange		{ $$ = $2;
								  OpNode nd[] = { $1 };
								  $2->textLengthMax($1->textLength());
								  /* register text as 3rd child */
								  $2->addChildren(NUM_OF(nd), nd); }
		| text PLUS text		{ $$ = new StrConcat($1, $3);
								  $$->textLengthMax($1->textLength() +
													$3->textLength()); }
		| '(' text ')'			{ $$ = $2; }
		| SLITERAL
		| SIDENT

pointlist: pointlist ',' arith ',' arith
								{ $$ = $1;
								  $1->addPoint($3, $5); };
		| arith ',' arith		{ $$ = new NodePointList;
								  $$->addPoint($1, $3); }

intervallist: intervallist ',' intervalspec
								{ $$ = $1;
								  OpNode nd[] = { $3 };
								  $1->addChildren(NUM_OF(nd), nd); }
		| intervalspec			{ $$ = new InInterval;
								  OpNode nd[] = { $1 };
								  $$->addChildren(NUM_OF(nd), nd); }

intervalspec: ':'						{ $$ = new IntTrue; }
			| '[' ':' ']'				{ $$ = new IntTrue; }
			| '(' ':' ']'				{ $$ = new IntTrue; }
			| '[' ':' ')'				{ $$ = new IntTrue; }
			| '(' ':' ')'				{ $$ = new IntTrue; }
			| arith						{ $$ = new IntEq($1); }
			| '[' arith ']'				{ $$ = new IntEq($2); }
			| arith ':'					{ $$ = new IntLe($1); }
			| '[' arith ':' ']'			{ $$ = new IntLe($2); }
			| '[' arith ':' ')'			{ $$ = new IntLe($2); }
			| '(' arith ':' ']'			{ $$ = new IntLt($2); }
			| '(' arith ':' ')'			{ $$ = new IntLt($2); }
			| ':' arith					{ $$ = new IntGe($2); }
			| '[' ':' arith ']'			{ $$ = new IntGe($3); }
			| '[' ':' arith ')'			{ $$ = new IntGt($3); }
			| '(' ':' arith ']'			{ $$ = new IntGe($3); }
			| '(' ':' arith ')'			{ $$ = new IntGt($3); }
			| arith ':' arith			{ $$ = new IntCC($1, $3); }
			| '[' arith ':' arith ']'	{ $$ = new IntCC($2, $4); }
			| '(' arith ':' arith ']'	{ $$ = new IntOC($2, $4); }
			| '[' arith ':' arith ')'	{ $$ = new IntCO($2, $4); }
			| '(' arith ':' arith ')'	{ $$ = new IntOO($2, $4); }

%%

}
