//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Selector.cc: --- implementation of selector methods
//
//	DESCRIPTION:
//		Implemenation of methods in OperatorNode + SymbolTable
//
//	CHANGE HISTORY:
//		UL, 1998-05-15: 1.0.0: - first version
//		UL, 2000-07-03: 1.1.0: - added StringLiteral() constructor
//
//	AUTHOR:
//		U. Lammers (UL), Fri May 15 09:37:22 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/Selector.cc,v 1.4 2002/12/17 17:48:11 jbakker Exp $
//

#include "basicdefs.h"
#include "Selector.h"
#include "SelectError.h"

using namespace SelectError;

#include <algorithm>

//
//	default arithmetic method for operator node
//
static inline double returnValue(OpNode me)
{
	return me->value();
} // returnValue()

//
//	default logical method for operator node
//
static inline bool returnFalse(OpNode)
{
	return false;
} // returnFalse()

//
//	default string method for operator node
//
static inline char *returnText(OpNode me)
{
	return me->text();
} // returnText();


OperatorNode::OperatorNode()
{
	//	initialize data members
	_isNumeric	= true;
	_isConst	= true;
	_value		= 0.;
	_text		= 0;
	_textLength	= 0;
	_nChildren	= 0;
	_refCount	= 1;
	arith		= returnValue;
	logic		= returnFalse;
	textual		= returnText;
} // OperatorNode::OperatorNode()


void OperatorNode::finalize()
{
} // OperatorNode::finalize()


void OperatorNode::registerLogicFunc(LogicFunc logicFunc)
{
	logic = logicFunc;
} // OperatorNode::registerLogicalFunc()


void OperatorNode::registerLogicFuncArg(LogicFuncArg logicFuncArg)
{
	logicArg = logicFuncArg;
} // OperatorNode::registerLogicalFuncArg()


void OperatorNode::registerArithFunc(ArithFunc arithFunc)
{
	arith = arithFunc;
} // OperatorNode::registerArithFunc()


void OperatorNode::registerTextFunc(TextFunc textFunc)
{
	textual = textFunc;
} // OperatorNode::registerTextFunc()


unsigned OperatorNode::refCount() const
{
	return _refCount;
} // OperatorNode::refCount()


void OperatorNode::incrRefCount()
{
	++_refCount;
} // OperatorNode::incrRefCount()


void OperatorNode::decrRefCount()
{
	_refCount = std::max(INT(0), INT(_refCount) - 1);
} // OperatorNode::decrRefCount()


void OperatorNode::value(double val)
{
	_value		= val;
	_isNumeric	= true;
} // OperatorNode::value()


void OperatorNode::text(const char *s)
{
	_isNumeric = false;
	delete [] _text;
	_textLength = strlen(s);
	_text = new char[_textLength+1]; _text[_textLength] = '\0';
	std::strcpy(_text, s);
} // OperatorNode::text()


void OperatorNode::textLengthMax(unsigned len)
{
	_isNumeric = false;
	delete [] _text;
	_textLength = len;
	_text = new char[_textLength+1]; _text[_textLength] = '\0';
} // OperatorNode::textLengthMax()


OperatorNode::~OperatorNode()
{
	//	delete children
	for (unsigned n=0; n<_nChildren; ++n)
		if (!isIdentifierNode())	// identifiers are deleted in
									// symbols table destructor
			if (_child[n]->refCount() > 1)
				_child[n]->decrRefCount();
									//	decrement reference count
			else
				delete _child[n];	//	this is the last reference - the
									//	node can really be deleted
	//	delete text string
	delete [] _text;
} // OperatorNode::~OperatorNode()


void OperatorNode::addChildren(unsigned nChildren, const OpNode *child)
{
	if (nChildren + _nChildren > MAX_CHILDREN) {
		selerrstr<<"OperatorNode::addChildren: Warning: Maximum number "
				   "of children ("<<MAX_CHILDREN<<") exceeded: "
				 <<nChildren + _nChildren
				 <<" - can only add "<<MAX_CHILDREN-_nChildren;
		SELTHROW(SelectException);
	}

	unsigned nAdd = std::min(nChildren, MAX_CHILDREN-_nChildren);
	for (unsigned n=0; n<nAdd; ++n) {
		_child[n+_nChildren] = child[n];
		child[n]->incrRefCount();	//	notify child that it has been
									//	registered
		_isConst = _isConst && child[n]->isConst();
									//	node is constant if all of its
									//	children are constant
	}

	_nChildren += nAdd;
} // OperatorNode::registerChildren()


//
//	-----------------( identifiers and symbol table )------------------------
//
OpNode SymbolTable::identifier(const std::string &name)
{
	unsigned nSyms = this->size();
	for (unsigned n=0; n<nSyms; ++n)
		if ((*this)[n]->name() == name)
			//	symbol found
			return (*this)[n];

	//	a new symbol - add to the end
	Identifier *id = new Identifier(name);
	this->push_back(id);

	return id;
} // SymbolTable::identifier()


SymbolTable::~SymbolTable()
{
	for (unsigned n=0; n<this->size(); ++n)
		delete (*this)[n];
} // SymbolTable::~SymbolTable()


StringLiteral::StringLiteral(const char *s)
{
	std::string tmp(s);
	std::string::size_type pos = tmp.find("\\\"");
	while(pos != std::string::npos) {
		tmp.erase(pos, 1);
		pos = tmp.find("\\\"");
	}

	text(tmp.c_str());
} // StringLiteral::StringLiteral()
