//---------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Selector.h:	--- central include file of core selector library ---
//
//	DESCRIPTION:
//		This is the central include file of the core selector library.
//		It defines the structure of the internal tree to be dynamically
//		built during the expression parsing.
//		The class OperatorNode above contains to pointers to functions
//		to evalyate the node in an arithmetic or logical context.
//		It would be cleaner to make these virtual functions, however,
//		this was sacrified for speed [the implementation with virtual
//		function was about 25% slower]
//
//	CHANGE HISTORY:
//		UL, 1998-05-01: 1.0.0: - first version
//		UL, 1998-05-19: 2.0.0: - overhaul for selectlib v2.0
//		UL, 1999-04-16: 2.1.0: - changed class name Node -> OpNode to
//								 resolve name conflict with Cal
//		UL, 2000-07-03: 2.2.0: - several changed related to support of
//								 string columns
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug 27 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/Selector.h,v 1.3 2002/05/24 16:00:10 ulammers Exp $
//

#ifndef SELECTOR_H
#define SELECTOR_H

//#include "Dal.h"
#include "TableDescriptor.h"

#include <cassert>
#include <string>
#include <vector>

class OperatorNode;					// forward declaration
class Evaluator;					// forward declaration
class SelectWCS;					// forward declaration
class Evaluator;					// forward declaration
typedef OperatorNode *OpNode;		// a OpNode is a pointer to an operator
typedef bool (*LogicFunc)(OpNode);	// logical evaluation function
typedef bool (*LogicFuncArg)(OpNode, double);
									// logical evaluation function (has to
									// be given one argument
typedef double (*ArithFunc)(OpNode);// arithmetic evaluation function
typedef char * (*TextFunc)(OpNode); // string evaluation function

class OperatorNode {
public:
	friend class Evaluator;			//	need to get pointer to values

	OperatorNode();
		// constructor - initializes data members

	void addChildren(unsigned nChildren, const OpNode child[]);

	virtual void finalize();
		//	finalize node construction - this provides the hook to
		//	do optimizations etc.

	void registerLogicFunc(LogicFunc logicFunc);
	void registerLogicFuncArg(LogicFuncArg logicFuncArg);
	void registerArithFunc(ArithFunc arithFunc);
	void registerTextFunc(TextFunc textFunc);
		// register children/evaluation functions

	ArithFunc		arith;
		//	evalulate node in arithmetic context

	LogicFunc		logic;
	LogicFuncArg	logicArg;
		//	evalulate node in logical context

	TextFunc		textual;
		//	evaluation in textual context

	inline bool		isNumeric() const { return _isNumeric; }

	inline double	value()	const { return _value; }
		//	return value of the node if it is terminal, i.e. a constant,
		//	or identifier

	virtual bool isConst() const { return _isConst; }
		//	a node is constant if all its children are constant terminals

	inline char *text() const { return _text; }
		//	return pointer to textual value if it is terminal
	inline unsigned textLength() const { return _textLength; }
	void textLengthMax(unsigned);
		//	return/set length of text

	unsigned		nChildren() const { return _nChildren; }
	inline OpNode	child(unsigned which = 0) const { return _child[which]; }
		//	access the children

	static unsigned const MAX_CHILDREN = 16;	// allow max 16 subtrees

	void incrRefCount();
	void decrRefCount();
	unsigned refCount() const;

	virtual void addPoint(const OpNode, const OpNode) {}
	virtual bool isIdentifierNode() const { return false; }
	virtual const std::string &name() const { static std::string s="";
											  return s; }
	virtual unsigned *arrayData() const { return 0; }
	virtual unsigned auxData(unsigned) const { return 0; }
	virtual const SelectWCS *maskWCS(unsigned) const { assert(0); return 0; }
	virtual const SelectWCS *eventTableWCS(unsigned) const { assert(0);
															 return 0; }
		//	to be overridden in derived classes

	virtual ~OperatorNode();
		//	destructor - frees up memory

protected:
	//	set value/text field
	void value(double);
	void text(const char *);
	void setConstness(bool constness) { _isConst = constness; }

private:
	unsigned	_refCount;	//	How often is this node used?
	unsigned	_nChildren;	//	How many children?
	OpNode		_child[MAX_CHILDREN];

	//	for numeric nodes
	bool		_isNumeric;	//	Is this a numeric node?
	bool		_isConst;	//	Is the node constant?
	double		_value;		//	What's its numeric value?

	//
	char		*_text;			//	What's its string value?
	unsigned	_textLength;	//	maximum length
};


//
//	------------------( identifier class )---------------------------------
//
class Identifier : public OperatorNode {
public:
	Identifier(const std::string &name = "") : _name(name)
			{ setConstness(false); }

	const std::string &name() const { return _name; }
	void name(const std::string &name) { _name = name; }
	bool isIdentifierNode() const { return true; }
	bool isConst() const { return false; }

	~Identifier() {}

private:
	std::string _name;
};


//
//	--------------------------( symbol table )------------------------------
//
const std::string ROWNUMBER_ID	= "ROWNUMBER";	// id for the row number
const std::string SELECT_COLUMN	= "SELECTCOLUMN";
											// id for column holding selection
											// status
const std::string SELECT_MASK	= "SELECTMASK";// mask for selection

class SymbolTable : public std::vector<Identifier *> {
public:
	SymbolTable() {}
	OpNode identifier(const std::string &name);
	~SymbolTable();
};

//
//	a numeric/string literal (constant)
//
class NumericLiteral : public OperatorNode {
public:
	NumericLiteral(double val) { value(val); setConstness(true); }
	~NumericLiteral() {}
};

class StringLiteral : public OperatorNode {
public:
	StringLiteral(const char *s);
	~StringLiteral() {}
};

typedef struct SelectorGlobals {
	SymbolTable		*symbolTable;	//	pointer to symbol table
	OpNode			selExpression;	//	the final expression tree
	bool			isSelExprLogic;	//	can tree be eval. in logical context?
	const TableDescriptor *tableDescriptor;// table descriptor
	Evaluator		*evaluator;		//	the current evaluator
};

extern SelectorGlobals	*selectorGlobals;
extern SelectorGlobals	selectorGlobalsInstance;

#endif
