//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010, European Space Agency
//
//---------------------------------------------------------------------------
//
//	ExprColumnGenerator.cc:	--- implementation of ExprColumnGenerator class
//
//	CHANGE HISTORY:
//		UL, 1998-05-19: 1.0.0: - first version
//		UL, 1998-11-02: 1.2.0: - handle case that newly to be generated
//								 column exists already: when this happens:
//									o generate new column with temporary
//									  name
//									o delete original column
//									o rename new column
//		UL, 1999-03-03: 1.3.0: - changes enforced by Lowmem Dal
//		UL, 1999-08-06: 1.4.0: - fix in ExprColumnGenerator::doIt: usage
//								 of non-existent columns in expression
//								 was not caught
//		UL, 1999-11-03: 1.5.0: - added support for new unsigned types
//								 in dal>=1.80
//		UL, 2000-10-25: 1.5.1: - release of column data objects after last
//								 use
//
//	AUTHOR:
//		U. Lammers (UL), Tue May 19 15:05:04 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/ExprColumnGenerator.cc,v 1.3 2002/07/19 10:32:35 ulammers Exp $
//

#include "../selcore/basicdefs.h"
#include "DataFilter.h"
#include "ExprColumnGenerator.h"
#include "ServiceUtils.h"

#if !defined(NON_SAS)
#include "SelectError.h"

using namespace SelectError;

//	construct filter with boolean expression
ExprColumnGenerator::ExprColumnGenerator(const std::string &expression,
										 const std::string &columnName,
										 Dal::DataType dataType)
		: ColumnGenerator(columnName, dataType),
		  _expression(expression), _dalTableDescriptor(0), _evaluator(0)
{
} // ExprColumnGenerator::ExprColumnGenerator()


template<class DalType>
static void doItReally(const Evaluator *evaluator,
					   const Table &table,
					   double **&valPtr,
					   unsigned &nNumericIds, void **&dataVec,
					   ServiceUtils::ConvFuncPtr *&convFuncVec,
					   char **&textValPtr, unsigned &nTextIds,
					   char **&textDataVec, unsigned *&textFieldLength,
					   DalType *newColumn)
{
	unsigned nRows=table.rows();
	for (unsigned n=0; n<nRows; ++n) {
		//	assign variable values
		for (unsigned i=0; i<nNumericIds; ++i)
			*valPtr[i] = (*convFuncVec[i])(dataVec[i], n);

		//	assign row number
		*valPtr[nNumericIds] = DBL(n+1);

		//	assign string variable values
		for (unsigned i=0; i<nTextIds; ++i)
			strncpy(textValPtr[i], textDataVec[i], textFieldLength[i]);	

		//	evaluate expression
		newColumn[n] = static_cast<DalType>(evaluator->evalArith());
	}
}


void ExprColumnGenerator::doIt(const Table &inTable, Table &outTable) const
{
	//	retarted instantiation of Evaluator object
	delete _evaluator;
	delete _dalTableDescriptor;
	_dalTableDescriptor = new DalTableDescriptor(&inTable);
	_expression = DataFilter::expandSelectionExpression(_expression, inTable);
	_evaluator = new Evaluator(_expression, _dalTableDescriptor);

	//	create the new column with a temporary name first and rename it
	//	later on; this way we can cope with expressions like
	//	ACOL = ACOL + BCOL
	//
	std::string	tmpColName("___tmp___");

	//	add new column
	std::vector<unsigned long> s; s.push_back(1);
	Column *newCol = outTable.addColumn(tmpColName, dataType(), _expression,
										"", s);

	//	prepare evaluation
	double						**valPtr;
	void						**dataVec;
	ServiceUtils::ConvFuncPtr	*convFuncVec;
	unsigned					nNumericIds;

	char						**textValPtr;
	unsigned					nTextIds;
	char						**textDataVec;
	unsigned					*textFieldLength;

	std::vector<ColumnData *> cdv =
		ServiceUtils::prepareEvaluation(_evaluator, inTable, valPtr,
										nNumericIds, dataVec, convFuncVec,
										textValPtr, nTextIds, textDataVec,
										textFieldLength);

	#define	DOIT_FOR(typeTag, type)										\
		case typeTag: {	ColumnData *cd = newCol->data();				\
						type *ptr = cd->type##Data();					\
						doItReally(_evaluator, inTable, valPtr,			\
								   nNumericIds,	dataVec, convFuncVec,	\
								   textValPtr, nTextIds, textDataVec,	\
								   textFieldLength, ptr);				\
						delete cd;										\
						break; }

	switch (dataType()) {
		DOIT_FOR(Dal::Int8, int8);
		DOIT_FOR(Dal::Int16, int16);
		DOIT_FOR(Dal::Uint16, uint16);
		DOIT_FOR(Dal::Int32, int32);
		DOIT_FOR(Dal::Uint32, uint32);
		DOIT_FOR(Dal::Real32, real32);
		DOIT_FOR(Dal::Real64, real64);
		default:
			selerrstr<<"Cannot convert to specified output type "
					 <<UNT(dataType())<<selerror("UnsupportedType");
			SELSTDTHROW;
			break;
	}

	//
	//	be nice and free-up memory
	//
	delete [] valPtr;
	delete [] dataVec;
	delete [] convFuncVec;
	delete [] textValPtr;
	delete [] textDataVec;
	delete [] textFieldLength;
	unsigned ncdv = cdv.size();
	for (unsigned n=0; n<ncdv; delete cdv[n++]);

	//	check if the new column exists alreay - if yes, give a warning
	if (outTable.hasColumn(columnName())) {
		selerrstr<<"The new column `"<<columnName()<<"' exists "
				   "already in the table "<<outTable.qualifiedName()
				 <<" - it will get overwritten"
				 <<selwarning("OverwriteColumn");

		outTable.deleteColumn(columnName());
	}

	newCol->rename(columnName());
} // ExprColumnGenerator::doIt()

ExprColumnGenerator::~ExprColumnGenerator()
{
	delete _dalTableDescriptor;
	delete _evaluator;
} // ExprColumnGenerator::~ExprColumnGenerator()

#endif
