//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ServiceUtils.cc: --- implementation of selector service utilities
//
//	DESCRIPTION:
//		A collection of utilities for the selector services
//
//	CHANGE HISTORY:
//		UL, 1998-05-19: 1.0.0: - first version
//		UL, 1999-03-03: 1.0.1: - changes enforced by Lowmem Dal
//		UL, 1999-05-12: 1.0.2: - added implementation of
//								 attributeFreeExpression()
//		UL, 1999-06-03: 1.1.0: - bug fixes in attributeFreeExpression()
//								   + reference to non-existing attributes did
//								     not lead to an error
//								   + _'s in attributes names was not allowed
//		UL, 1999-06-28: 1.2.0: - better error messages
//		UL, 1999-09-13: 1.3.0: - removed interim implementation of fconvint8
//								 after fixing of DAL SPR 589: incorrect
//								 treatment of FITS B-type
//		UL, 1999-10-08: 1.3.1: - attributeFreeExpression accepts optional
//								 argument with list of special attribute
//								 names which must not be replaced
//		UL, 1999-11-03: 1.4.0: - added support for new unsigned types
//								 in dal>=1.80
//		UL, 2000-07-03: 1.5.0: - several changes related to support of
//								 string columns
//		UL, 2003-02-27: 1.6.0: - fixed bug in findNextAttribute()
//		UL, 2003-05-16: 1.7.0: - fixed bug in prepareEvaluation() - case of
//								 reference to non-existing string column
//								 was not not caught
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug 27 11:48:12 MET DST 1997
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/ServiceUtils.cc,v 1.5 2003/05/19 12:24:44 ulammers Exp $
//

#include <algorithm>

#include "DataFilter.h"
#include "ServiceUtils.h"
#include "SelectError.h"

#define fconv(type)									\
inline double fconv##type(type *v, unsigned long i)	\
{													\
	return static_cast<double>(v[i]);				\
}

fconv(bool8);
fconv(int8);
fconv(int16);
fconv(uint16);
fconv(int32);
fconv(uint32);
fconv(real32);
fconv(real64);

//
//	prepareEvaluation: assign attribute value, figure list of identifiers
//
std::vector<ColumnData *>
	ServiceUtils::prepareEvaluation(const Evaluator *evaluator,
									const Table &table,
									double **&valPtr,
									unsigned &nNumericIds, void **&dataVec,
									ConvFuncPtr *&convFuncVec,
									char **&textValPtr,
									unsigned &nTextIds, char **&textDataVec,
									unsigned *&textFieldLength)
{
	//	vector with to be deleted column data vector
	std::vector<ColumnData *> cdv;

	//	check if flag column name and bit position is given
	std::string flagColumnName(table.hasAttribute(DataFilter::FLAGCOLUMN)
						? table.attribute(DataFilter::FLAGCOLUMN)->asString()
						: std::string(""));
	unsigned short flagBit = table.hasAttribute(DataFilter::FLAGBIT)
						? table.attribute(DataFilter::FLAGBIT)->asInt() : 0;
	evaluator->setFlagData(flagColumnName, flagBit);

	//
	//	What numeric symbols are used in the expression?
	//
	NumericSymbols idents = evaluator->numericIdentifiers();
	nNumericIds = idents.name.size();
	valPtr = new (double *)[nNumericIds+1];
										// last element is used to hold pointer
										// to row number
	static double rowNr;				// dummy row number
	valPtr[nNumericIds] = evaluator->rowNumber();
	if (!valPtr[nNumericIds])
		//	row number is not specified in expression - make the last
		//	entry in valPtr point to some dummy variable
		valPtr[nNumericIds] = &rowNr;

	for (unsigned n=0; n<nNumericIds; ++n)
		valPtr[n] = idents.valuePtr[n];

	//	set up vector with pointers to type conversion functions
	//	this is necessary because the Dal does not provide automatic
	//	type conversion
    convFuncVec = new ConvFuncPtr[nNumericIds];
	dataVec = new (void *)[nNumericIds];

	for (unsigned i=0; i<nNumericIds; ++i)
		if (table.hasColumn(idents.name[i])) {
			const Column *col = table.column(idents.name[i]);
			ColumnData *cd = col->data();
			switch (col->dataType()) {
				case Data::Bool:
						dataVec[i] = cd->boolData();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvbool8);
						break;
				case Data::Int8:
						dataVec[i] = cd->int8Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvint8);
						break;
				case Data::Int16:
						dataVec[i] = cd->int16Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvint16);
						break;
				case Data::Uint16:
						dataVec[i] = cd->uint16Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvuint16);
						break;
				case Data::Int32:
						dataVec[i] = cd->int32Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvint32);
						break;
				case Data::Uint32:
						dataVec[i] = cd->uint32Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvuint32);
						break;
				case Data::Real32:
						dataVec[i] = cd->real32Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvreal32);
						break;
				case Data::Real64:
						dataVec[i] = cd->real64Data();
						convFuncVec[i] =
									reinterpret_cast<ConvFuncPtr>(fconvreal64);
						break;
				default:
						selerrstr<<"Column `"<<col->qualifiedName()
								 <<"' is not numeric"
								 <<selerror("NonNumericColumnUnknownType");
						SELSTDTHROW;
						break;
			}
			cdv.push_back(cd);
		} else {
			selerrstr<<"Column `"<<idents.name[i]<<"' does not exist in "
					   "table `"<<table.qualifiedName()<<"'"
					 <<selerror("UnknownColumn");
			SELSTDTHROW;
		}

	//
	//	What text symbols are used in the expression?
	//
	StringSymbols sidents = evaluator->stringIdentifiers();
	nTextIds = sidents.name.size();
	textValPtr = new (char *)[nTextIds];
	textDataVec = new (char *)[nTextIds];
	textFieldLength = new unsigned[nTextIds];
	for (unsigned n=0; n<nTextIds; ++n) {
		textValPtr[n] = sidents.textPtr[n];
		textFieldLength[n] = sidents.maxLength[n];
		if (table.hasColumn(sidents.name[n])) {
			const Column *col = table.column(sidents.name[n]);
			ColumnData *cd = col->data();
			if (col->dataType() == Data::DString) {
				textDataVec[n] = cd->stringData();
				cdv.push_back(cd);
			} else {
				selerrstr<<"Column `"<<col->qualifiedName()<<"' is not a "
						   "string column"
						 <<selerror("NonStringColumn");
				SELSTDTHROW;
			}
		} else {
			selerrstr<<"Column `"<<sidents.name[n]<<"' does not exist in "
					   "table `"<<table.qualifiedName()<<"'"
					 <<selerror("UnknownColumn");
			SELSTDTHROW;
		}
	}

	return cdv;
} // ServiceUtils::prepareEvaluation()

struct AttString { std::string::size_type pos; unsigned len; };
static AttString findNextAttribute(const std::string &s, unsigned i0 = 0)
{
	AttString a = { std::string::npos, 0 };
	unsigned l = s.length(), i = i0;
	while (i < l)
		if (s[i++] == '#') {
			//
			//	begining of attribute reference
			//
			a.pos = i-1;
			unsigned j = i;
			while (j<l && (isalnum(s[j]) || s[j] == '_' || s[j] == '-'))
				++j;
			a.len = j - a.pos;
			break;
		}

	return a;
} // findNextAttribute()


std::string ServiceUtils::attributeFreeExpression(
											const std::string &expression,
											const Table *table,
											const std::vector<std::string>
													*specialAttributes)
{
	std::string e(expression);
	std::vector<std::string> empty;
	const std::vector<std::string> &specAttr =
					specialAttributes ? *specialAttributes : empty;

	unsigned i0 = 0;
	AttString a = findNextAttribute(e, i0);
	while(a.len) {
		//
		//	check if the attribute exists
		//
		std::string att(e.substr(a.pos+1, a.len-1));
		if (find(specAttr.begin(), specAttr.end(), att) != specAttr.end()){
			//
			//	special attribute - ignore it
			//
			i0 = a.pos+1;
			a = findNextAttribute(e, i0);
			continue;
		}

		if (table->hasAttribute(att))
			//
			//	yes, get the attribute value as a string and replace the
			//	reference with this value
			//
			e.replace(a.pos, a.len, table->attribute(att)->asString());

		i0 = a.pos+1;
		a = findNextAttribute(e, i0);
	}

	return e;
} // ServiceUtils::attributeFreeExpression()
