//----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ExprTranslator.h: --- utility class to construct selectlib expressions
//
//	DESCRIPTION:
//		The class ExprTranslator provides two static methods to convert
//		region specification from imgdisplay and dsplot into selectlib
//		compliant expressions.
//
//	CHANGE HISTORY:
//		UL, 1999-02-17: 1.0.0: - first version
//		UL, 1999-09-20: 2.0.0: - taking binning factors !=1 into account
//							   - added two distinct classes for
//								 ImgDisplay and Dsplot regions
//		UL, 1999-10-25: 2.1.0: - added support for 2D regions from Dsplot
//
//	AUTHOR:
//		U. Lammers (UL), Wed Feb 17 13:22:25 MET 1999
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/ExprTranslator.h,v 1.2 2001/08/10 16:15:00 ulammers Exp $
//

#ifndef EXPR_TRANSLATOR_H
#define EXPR_TRANSLATOR_H

#include <string>
#include "errstr.h"

class RegionSpecification {
public:
	//	exception classes
	class InvalidRegionSpec {};
	class InvalidType : public InvalidRegionSpec {};

	RegionSpecification(const std::string &spec) : _spec(spec) { }
		//
		//	PURPOSE:
		//		Create a RegionSpecification object from a string
		//

	const std::string &spec() const { return _spec; }
		//
		//	PURPOSE:
		//		Return region specification the object was instantiated with
		//


	virtual std::string toSelectlibExpression(const std::string &xcolName,
											  const std::string &ycolName,
											  double binFactorX = 1.,
											  double binFactorY = 1.,
											  double minX = 1.,
											  double minY = 1.) const = 0;
		//
		//	PURPOSE:
		//		convert a region specification from imgdisplay
		//		into a selection expression which uses a region filter;
		//		binning factors are considered by multiplication of
		//		all dimensions and necessary translations of region filters,
		//		e.g. circle -> ellipse for binFactorX != binFactorY
		//


	virtual std::string toSelectlibExpression(const std::string &colName)
	const = 0;
		//
		//	PURPOSE:
		//		Convert a region specification from dsplot
		//		into a selection expression
		//


	virtual ~RegionSpecification() { }

protected:
	std::string	_spec;
}; // RegionSpecification



class ImgdisplayRegionSpecification : public RegionSpecification {
public:
	ImgdisplayRegionSpecification(const std::string &spec)
		: RegionSpecification(spec) { }

	std::string toSelectlibExpression(const std::string &xcolName,
									  const std::string &ycolName,
									  double binFactorX = 1.,
									  double binFactorY = 1.,
									  double minX = 1.,
									  double minY = 1.) const;

	std::string toSelectlibExpression(const std::string &) const
	{ errstr<<"ImgdisplayRegionSpecification object cannot convert "
				   "dsplot regions"<<Sas::error("NoSuchOperation");
	  return std::string();
	}

	~ImgdisplayRegionSpecification() { }

private:
	std::string circleArgs(const std::string &,
						   double, double, double, double) const;
	std::string ellipseArgs(const std::string &,
							double, double, double, double) const;
	std::string boxArgs(const std::string &,
						double, double, double, double) const;
	std::string polygonArgs(const std::string &,
							double, double, double, double) const;
	std::string pointArgs(const std::string &,
						  double, double, double, double) const;

	std::string convertPosition(const std::string &, double, double,
							    const std::string &, double, double) const;
	std::string convertDimension(const std::string &, double) const;
}; // ImgdisplayRegionSpecification



class DsplotRegionSpecification : public RegionSpecification {
public:
	DsplotRegionSpecification(const std::string &spec)
		: RegionSpecification(spec) { }

	std::string toSelectlibExpression(const std::string &xColName,
									  const std::string &yColName,
									  double = 1.,
									  double = 1.,
									  double = 1.,
									  double = 1.) const;

	std::string toSelectlibExpression(const std::string &colName) const;

	~DsplotRegionSpecification() { }

private:
	enum RegionType { NONE = -1, ABOVE_LINE, BELOW_LINE, LEFT_OF_LINE,
					  RIGHT_OF_LINE, INSIDE_POLY, OUTSIDE_POLY };

	RegionType type(const std::vector<std::string> &comp) const;
}; // DsplotRegionSpecification

#endif



