//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Gti.h:	--- Good Time Intervals class and lists of Good-Time-Intervals
//
//	DESCRIPTION:
//		An instance of class <Gti> represents a Good Time Interval defined
//		via a lower and upper time boundary. <GtiList> is a time-ordered
//		list of <Gti>s.
//
//	CHANGE HISTORY:
//		UL, 1998-02-26: 1.0.0: - first version
//		UL, 1999-10-21: 1.1.0: - added Gti::ETERNITY
//							   - added GtiList::operator|=	(logical or)
//							   - added GtiList::operator&=	(logical and)
//		UL, 2000-09-04: 1.1.1: - fixed typo in operator declaration
//
//	AUTHOR:
//		U. Lammers (UL), Thu Feb 26 15:10:16 MET 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/Gti.h,v 1.5 2002/12/17 17:48:11 jbakker Exp $
//

#ifndef GTI_H
#define GTI_H

#include <iosfwd>
#include <vector>
#include <string>

class Gti {
public:
	static const double ETERNITY = 1.e99;

	//	construct from explicit Gti boundaries
	//	In:
	//		start: lower GTI boundary
	//		stop : upper GTI boundary
	Gti(double start = 0., double stop = ETERNITY);

	double start() const { return _start; }
	void setStart(double start) { _start = start; }
	double stop() const { return _stop; }
	void setStop(double stop) { _stop = stop; }
	double length() const { return stop() - start(); }

	//	return true if specified time is within GTI
	bool inside(double time) const;

	~Gti() {};
private:
	double	_start, _stop;	// always _start <= _stop
};

//
//	print a GTI to a stream
//
std::ostream &operator<<(std::ostream &os, const Gti &gti);


class GtiList : public std::vector<Gti> {
public:
	//	default constructor - create empty list; if <allGood> is set to
	//	true the list contains a single GTI going from t=0 to t=+Inf
	GtiList(bool allGood = false);

	//	destructor
	~GtiList() {}

	//	check if a time is within one set - return pointer to corresponding
	//	GTI is yes, NULL otherwise
	const Gti *inside(double time) const;

	//	sort the list such that
	//	t_i_start <= t_i+1_start forall i=0, this->size()-1
	void sortIt();

	//	combine list with second list (logical OR)
	//
	//	+----------------+				GTI 1
	//				+--------------+	GTI 2
	//			||
	//			\/
	//	+--------------------------+	combined
	GtiList& operator|=(const GtiList &list);

	//	combine list with second list (logical AND)
	//	+----------------+				GTI 1
	//				+--------------+	GTI 2
	//			||
	//			\/
	//				+----+				combined
	GtiList &operator&=(const GtiList &list);

	//	combine overlapping intervals (logical or)
	GtiList &reduce();

	//	compute ONTIME = sum of all GTI
	double ontime() const;

	enum State { UNKNOWN = 0, SORTED, REDUCED };
	bool sorted() const { return _state == SORTED || _state == REDUCED; }
	bool reduced() const { return _state == REDUCED; }

	const std::vector<std::string> &origins() const { return _origins; }
	void addOrigin(const std::string &orig) { _origins.push_back(orig); }
	void addOrigin(const std::vector<std::string> &orig) {
				_origins.insert(_origins.end(), orig.begin(), orig.end()); }

private:
	State						_state;
	std::vector<std::string>	_origins;
};

//
//	combine two GtiLists
//
GtiList operator|(const GtiList &lhs, const GtiList &rhs);
GtiList operator&(const GtiList &lhs, const GtiList &rhs);

//
//	print entire list of GTIs to a stream
//
std::ostream &operator<<(std::ostream &os, const GtiList &gtiList);
bool operator<(const Gti &gti1, const Gti &gti2);

#endif
