//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Region.h:	--- geometrical region class
//
//	DESCRIPTION:
//		An instance of class <Region> represents a geometrical area in
//		two dimensions; it consists of the intersection and union
//		of the following elemental shapes:
//			o point
//			o circle
//			o annulus
//			o ellipse
//			o elliptannulus
//			o rectangle
//			o rotrectangle
//			o box
//			o rotbox
//			o rhombus
//			o rotrhombus
//			o sector
//			o polygon
//
//	CHANGE HISTORY:
//		UL, 1998-03-25: 1.0.0: - first version
//		UL, 1998-04-06: 1.1.0: - added various shape types to bring in line
//								 with [1] v1998-03-10
//		UL, 1998-07-31: 1.2.0: - bug fix in RotRhombus constructor (needed
//								 to call Box<T> constructor) - credit to
//								 Tim Harris for spotting this
//		UL, 1999-11-17: 1.3.0: - added Shape<T>::include()/exclude()
//							     for changing the include/exclude status
//							     after construction
//		UL, 1999-12-17: 1.4.0: - renamed Region->GeoRegion to resolve name
//								 clash with Qt type name
//		UL, 2000-01-04: 1.4.1: - Shape::isLabeled() had no explicit return tye
//								 (SPR 1060: Fails on LInux (gcc-2.95.1))
//		UL, 2000-06-22: 1.5.0: - remove terminal vertex from Polgon if
//								 present
//		UL, 2001-02-05: 1.6.0: - added added Shape::contains()
//		UL, 2001-08-15: 1,7.0: - added Line::contains(): returns true
//								 if distance of test points to line is smaller
//								 than double epsilon
//		UL, 2002-01-28: 1.8.0: - added EllipticalAnnulus::toSelectExpression()
//								 (override base class method)
//		UL, 2002-07-10: 1.9.0: - minor mods to class Shape
//							   - Component is now a proper class
//							   - GeoRegion inherits from class Selecting
//		UL, 2002-11-26: 1.10.0:- added ShapeBasics::ShapeAttribute() (and
//								 corresponding setter/getter methods in
//								 Shape<T>) to support more user-friendly
//								 selection of background regions in ds9
//
//	AUTHOR:
//		U. Lammers (UL), Wed Mar 25 16:52:23 MET 1998
//
//	REFERENCES:
//		[1] FITS REGION Binary Table Design, ASC-FITS-REGION-1.0,
//			J. McDowell, A. Rots,
//			http://hea-www.harvard.edu/~arots/asc/fits/region.ps
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/Region.h,v 1.13 2003/06/26 12:17:08 gvacanti Exp $
//

#ifndef REGION_H
#define REGION_H

#include <iosfwd>
#include <string>
#include <vector>

#include "SelectError.h"

using namespace SelectError;

namespace ShapeUtils {
	template<class T>
	inline T SQR(T arg) { return arg*arg; }

	template<class T>
	inline T ABS(T arg) { return arg>=0 ? arg : -arg; }
};


namespace SelectlibRegion {

//
//	a point in 2-D; type of coordinates is template parameters
//
template<class T>
class Point {
public:
	//	construct point at (0/0)
	Point(T x = 0, T y = 0) : _x(x), _y(y) {}
	~Point() {}

	T x() const { return _x; }
	T y() const { return _y; }

	bool operator==(const Point<T> &p) const
			{ return p.x() == _x && p.y() == _y; }

private:
	T _x, _y;
}; // Point

template<class T>
std::ostream &operator<<(std::ostream &os, const Point<T> &p)
{
	os<<'('<<p.x()<<','<<p.y()<<')';

	return os;
} // operator()


template<class T>
bool operator<(const Point<T> &p1, const Point<T> &p2)
{
	if (p1.x() != p2.x())
		return p1.x() < p2.x();

	return p1.y() < p2.y();
}


template<class T>
class PointList : public std::vector<Point<T> > {
public:
	//	construct empty list
	PointList() {}

	//	construct from n x/y coordinates pairs
	PointList(const std::vector<T> &xCoords, const std::vector<T> &yCoords,
			  unsigned nPoints)
			: _simpleClosedPathOrder(false), _antiClockWiseOrder(false)
	{
	  for (unsigned n=0; n<nPoints; ++n)
			this->push_back(Point<T>(xCoords[n], yCoords[n]));
	}

	Point<T> centerOfGravity() const;

	void simpleClosedPath();
		//
		//	PURPOSE:
		//		Order the vertices such that the first one has the smallest
		//		X-coordinate and iterating from there to the last one
		//		visits all points exactly once in anti-clockwise order.
		//		Example:
		//                B
		//			E            D
		//				A     C
		//
		//		simpleClosedPath: E->A->C->D->B
		//

	void antiClockWiseOrder();
		//
		//	PURPOSE:
		//		Re-arrange the ordering of the points such that the
		//		first one has the smallest X/Y coordinates and iterating
		//		through the list progesses in anticlockwise order.
		//			

	bool hasIntersectingSegments() const;
		//
		//	RETURNS:
		//		true if at least two segments of the polygon resulting
		//		from joining adjacent points intersect
		//

	bool formsSimpleRectangle();
		//
		//	PURPOSE:
		//		Inspect wheather the list of points forms a simple rectangle
		//		whose edges are aligned with the axes; as a side effect
		//		the points are re-orderd with simpleClosedPath()
		//

	virtual ~PointList() {}

private:
	bool _simpleClosedPathOrder;
	bool _antiClockWiseOrder;
};

template<class T>
std::ostream &operator<<(std::ostream &os, const PointList<T> &pl)
{
	unsigned N = pl.size();
	os<<'[';
	for (unsigned n=0; n<N; ++n) {
		os<<pl[n];
		if (n<N-1)
			os<<',';
	}

	return os;
} // operator<<

class ShapeBasics {
public:
	enum ShapeType { POINT = 0, LINE, CIRCLE, ANNULUS, ELLIPSE, ELLIPTANNULUS,
					 SECTOR, BOX, ROTBOX, RECTANGLE, ROTRECTANGLE,
					 RHOMBUS, ROTRHOMBUS, POLYGON };
	enum ShapeColor { BLACK = 0, WHITE, RED, GREEN, BLUE, CYAN, MAGENTA,
					  YELLOW };
	enum ShapeAttribute { SOURCE = 0, BACKGROUND };

	static std::string toString(ShapeColor color);
	static ShapeColor toColor(const string &colorString);
};


class UnexpressionableRegion : public SelectError::SelectException {
public:
	UnexpressionableRegion(const std::string &s) : SelectError::SelectException(s){}
};

class Selecting {
public:
	Selecting() {}

	virtual std::string toSelectExpression(const std::string &xAxisName,
									  const std::string &yAxisName,
									  double binFactorX = 1.,
									  double binFactorY = 1.,
									  double minX = 1.,
									  double minY = 1.) const
			throw(UnexpressionableRegion);
		//
		//	PURPOSE:
		//		convert a 2-D region specification
		//		into a selection expression which uses a region filter;
		//		binning factors are considered by multiplication of
		//		all dimensions and necessary translations of region filters,
		//		e.g. circle -> ellipse for binFactorX != binFactorY
		//


	virtual std::string toSelectExpression(const std::string &axisName,
									  double bin = 1., double min = 0.) const
		throw(UnexpressionableRegion);
		//
		//	PURPOSE:
		//		Convert a 1D-region specification into a selection expression
		//

	virtual ~Selecting() {}

protected:
	std::string prefix(const std::string &xAxisName, const std::string &yAxisName,
				  bool included = true) const;
	std::string prefix(const std::string &axisName, bool included = true) const;
	std::string convertPosition(double x, double binX, double minX,
						   double y, double binY, double minY) const;
	std::string convertDimension(double dim, double f) const;
};
										

//
//	-----------------------( Shapes )----------------------------------------
//
template<class T>
class Shape : public virtual ShapeBasics, public virtual Selecting {
public:
	Shape() : _refPoint(), _included(true) {}
	Shape(bool included) : _refPoint(), _included(included) {}
	Shape(const Point<T> &refPoint, bool included = true,
		  const std::string &label = "", const ShapeColor color = WHITE,
			ShapeAttribute attrib = SOURCE) :
			_refPoint(refPoint), _included(included), _label(label),
			_color(color), _attrib(attrib) {}
	virtual ~Shape() {}

	Point<T> refPoint() const { return _refPoint; }
	bool	 included() const { return _included; }
	void	 include(bool included=true) { _included = included; }
	void	 exclude() { _included = false; }
	virtual	ShapeType type() const { return POINT; }
	virtual std::string name() const { return "POINT"; }
	virtual bool isLabeled() const { return !_label.empty(); }
	virtual void label(const std::string &l) { _label = l; }
	virtual const std::string &label() const { return _label; }
	virtual void color(const ShapeColor &color) { _color = color; }
	virtual ShapeColor color() const { return _color; }
	virtual void attribute(const ShapeAttribute &attrib) { _attrib = attrib; }
	virtual ShapeAttribute attribute() const { return _attrib; }
	virtual bool is2D() const { return true; }

	//	dummy versions - actual ones implemented in derived classes
	virtual std::vector<T> radii() const { return std::vector<T>(); }
	virtual std::vector<double> rotAngles() const
		{ return std::vector<double>(); }
	virtual Point<T> topRight() const	{ return Point<T>(); }
	virtual std::vector<T> pointsX() const	{ std::vector<T> x;
											  x.push_back(_refPoint.x());
											  return x;
											}
	virtual std::vector<T> pointsY() const	{ std::vector<T> y;
											  y.push_back(_refPoint.y());
											  return y;
											}

	virtual std::string toSelectExpression(const std::string &axisName,
									  double bin = 1., double min = 0.) const
			throw(UnexpressionableRegion);

	virtual std::string toSelectExpression(const std::string &xAxisName,
									  const std::string &yAxisName,
									  double binFactorX = 1.,
									  double binFactorY = 1.,
									  double minX = 1.,
									  double minY = 1.) const
			throw(UnexpressionableRegion);

	virtual bool contains(const Point<T> &) const {
		selerrstr<<"Generic shape does not implement point-in-shape test"
				 <<selerror("NotImplementedInBaseClass", 1);
		return false;
	}
	virtual bool dummy(int i) const { return i != 1; }
		//
		//	RETURN:
		//		True/false depending on whether the given point lies within
		//		(including the border) the shape or not.
		//

protected:
	Point<T>	_refPoint;	// reference point of this shape
	bool		_included;	// is the shape included or excluded
	std::string	_label;		// any text label
	ShapeColor	_color;		// color of the region (border)
	ShapeAttribute _attrib;	// attribute
}; // Shape

template<class T>
class RotatedShape : public virtual Shape<T> {
public:
	RotatedShape() : Shape<T>(), _rotAngle(0.) {}
	RotatedShape(const Point<T> &refPoint, double rotAngle = 0.,
				 bool included = true) :
		Shape<T>(refPoint, included), _rotAngle(rotAngle) {}
	~RotatedShape() {}

	double rotAngle() const { return _rotAngle; }
	std::vector<double> rotAngles() const { std::vector<double> a;
											a.push_back(_rotAngle);
											return a;
										  }

private:
	double _rotAngle;
};


template<class T>
class Line : public virtual Shape<T> {
public:
	Line() : Shape<T>(), _p2(Point<T>()), _length2(0.) {}
	//	referene point is first point
	Line(const Point<T> &refPoint, const Point<T> &p2,
		 bool included = true) :
		Shape<T>(refPoint, included), _p2(p2) {
			_length2 = ShapeUtils::SQR(p2.x()-refPoint.x()) +
					   ShapeUtils::SQR(p2.y()-refPoint.y());
		}
		//
		//	PURPOSE:
		//		Construct Line shape
		//
		//	ARGUMENTS:
		//		refPoint, p2	: end points
		//		included		: include status
		//

	~Line() {}

	Point<T> topRight() const { return _p2; }
	std::vector<T> pointsX() const { std::vector<T> x;
									 x.push_back(_refPoint.x());
									 x.push_back(_p2.x());
									 return x; }
	std::vector<T> pointsY() const { std::vector<T> y;
							 		 y.push_back(_refPoint.y());
									 y.push_back(_p2.y());
									 return y; }

	ShapeBasics::ShapeType type() const { return LINE; }
	std::string name() const { return "LINE"; }

	virtual bool is2D() const { return false; }

	std::string toSelectExpression(const std::string &axisName,
								   double bin = 1., double min = 0.) const
			throw(UnexpressionableRegion);

	inline bool contains(const Point<T> &p) const
	{
		if (_length2<=0.)
			return p == _p2;

		double deltaX = _p2.x() - refPoint().x(),
			   deltaY = _p2.y() - refPoint().y();
		//	see http://astronomy.swin.edu.au/pbourke/geometry/pointline/
		double u = ((p.x()-refPoint().x())*deltaX +
					(p.y()-refPoint().y())*deltaY)/_length2;
		if (u>=0 && u<=1) {
			//	compute square of distance between test point and intersection
			//	point with line segment
			//	consider point on line if l2 is smaller than sqrt(machine eps)
			return ShapeUtils::SQR(refPoint().x()+u*deltaX - p.x()) +
				   ShapeUtils::SQR(refPoint().y()+u*deltaY - p.y()) < 1.e-12;
		}

		return false;
	} // Line::contains()

private:
	Point<T>	_p2;
	double		_length2;
};


template<class T>
class Circle : public virtual Shape<T> {
public:
	Circle() : Shape<T>(), _radius(0) {}
	Circle(const Point<T> &refPoint, T radius = 0, bool included = true) :
		Shape<T>(refPoint, included), _radius(radius) {}
	~Circle() {}

	T radius() const { return _radius; }
	std::vector<T> radii() const
		{ std::vector<T> t; t.push_back(_radius); return t; }
	ShapeBasics::ShapeType type() const { return CIRCLE; }
	std::string name() const { return "CIRCLE"; }

	std::string toSelectExpression(const std::string &xAxisName,
							  const std::string &yAxisName,
							  double binFactorX = 1.,
							  double binFactorY = 1.,
							  double minX = 1.,
							  double minY = 1.) const
			throw(UnexpressionableRegion);

	inline bool contains(const Point<T> &p) const
	{
		bool incTest = ShapeUtils::SQR(p.x() - _refPoint.x()) +
					   ShapeUtils::SQR(p.y() - _refPoint.y())
						<= ShapeUtils::SQR(_radius);
		return included() ? incTest : !incTest;
	} // Circle::contains()

private:
	T _radius;
}; // Circle


template<class T>
class Annulus : public virtual Shape<T> {
public:
	Annulus() : Shape<T>(), _innerCircle(Circle<T>()),
						    _outerCircle(Circle<T>()) {}
	Annulus(const Point<T> &refPoint, T innerRadius = 0, T outerRadius = 0,
			bool included = true) :
			Shape<T>(refPoint, included),
			_innerCircle(refPoint, innerRadius, false),
			_outerCircle(refPoint, outerRadius, included) {}
	~Annulus() {}

	Circle<T> innerCircle() const { return _innerCircle; }
	Circle<T> outerCircle() const { return _outerCircle; }
	std::vector<T> radii() const { std::vector<T> t;
								   t.push_back(_innerCircle.radius());
								   t.push_back(_outerCircle.radius());
								   return t; }
	ShapeBasics::ShapeType type() const { return ANNULUS; }
	std::string name() const { return "ANNULUS"; }

	std::string toSelectExpression(const std::string &xAxisName,
							  const std::string &yAxisName,
							  double binFactorX = 1.,
							  double binFactorY = 1.,
							  double minX = 1.,
							  double minY = 1.) const
			throw(UnexpressionableRegion);

	inline bool contains(const Point<T> &p) const
	{
		bool incTest = _outerCircle.contains(p) &&
					   !_innerCircle.contains(p);
		return included() ? incTest : !incTest;
	} // contains()

private:
	Circle<T>	_innerCircle, _outerCircle;
}; // Annulus


template<class T>
class Ellipse : public virtual RotatedShape<T> {
public:
	Ellipse() : Shape<T>(), RotatedShape<T>(),
				_radiusMajor(0), _radiusMinor(0) {}
	Ellipse(const Point<T> &refPoint, T radiusMajor = 0, T radiusMinor = 0,
			double rotAngle = 0., bool included = true) :
			Shape<T>(refPoint, included),
			RotatedShape<T>(refPoint, rotAngle, included),
			_radiusMajor(radiusMajor), _radiusMinor(radiusMinor) {}
	~Ellipse() {}

	T radiusMajor() const { return _radiusMajor; }
	T radiusMinor() const { return _radiusMinor; }
	std::vector<T> radii() const { std::vector<T> t;
								   t.push_back(_radiusMajor);
								   t.push_back(_radiusMinor); return t; }

	ShapeBasics::ShapeType type() const { return ELLIPSE; }
	std::string name() const { return "ELLIPSE"; }

	std::string toSelectExpression(const std::string &xAxisName,
							  const std::string &yAxisName,
							  double binFactorX = 1.,
							  double binFactorY = 1.,
							  double minX = 1.,
							  double minY = 1.) const
			throw(UnexpressionableRegion);

private:
	T _radiusMajor, _radiusMinor;
};


template<class T>
class ElliptAnnulus : public virtual Shape<T> {
public:
	ElliptAnnulus() : Shape<T>(), _innerEllipse(Ellipse<T>()),
			_outerEllipse(Ellipse<T>()) {}
	ElliptAnnulus(const Point<T> &refPoint, T innerRadiusMajor = 0,
			T innerRadiusMinor = 0, T outerRadiusMajor = 0,
			T outerRadiusMinor = 0, double innerRotAngle = 0.,
			double outerRotAngle = 0., bool included = true) :
			Shape<T>(refPoint, included),
			_innerEllipse(refPoint, innerRadiusMajor, innerRadiusMinor,
						  innerRotAngle, false),
			_outerEllipse(refPoint, outerRadiusMajor, outerRadiusMinor,
						  outerRotAngle, included) {}
	~ElliptAnnulus() {}

	std::vector<T> radii() const { std::vector<T> t;
							  t.push_back(_innerEllipse.radiusMajor());
							  t.push_back(_innerEllipse.radiusMinor());
							  t.push_back(_outerEllipse.radiusMajor());
							  t.push_back(_outerEllipse.radiusMinor());
							  return t; }
	std::vector<double> rotAngles() const { std::vector<double> a;
									   a.push_back(_innerEllipse.rotAngle());
									   a.push_back(_outerEllipse.rotAngle());
									   return a; }


	Ellipse<T> innerEllipse() const { return _innerEllipse; }
	Ellipse<T> outerEllipse() const { return _outerEllipse; }
	ShapeBasics::ShapeType type() const { return ELLIPTANNULUS; }
	std::string name() const { return "ELLIPTANNULUS"; }

	std::string toSelectExpression(const std::string &xAxisName,
								   const std::string &yAxisName,
								   double binX,
								   double binY,
								   double minX,
								   double minY) const
			throw(UnexpressionableRegion);

private:
	Ellipse<T> _innerEllipse, _outerEllipse;
};


template<class T>
class Rectangle : public virtual Shape<T> {
public:
	Rectangle() : Shape<T>(), _topRight(Point<T>()) {}
	//	referene point is lower left corner
	Rectangle(const Point<T> &refPoint, const Point<T> &topRight,
			  bool included = true) :
		Shape<T>(refPoint, included), _topRight(topRight) {}
	~Rectangle() {}

	Point<T> topRight() const { return _topRight; }
	std::vector<T> pointsX() const { std::vector<T> x;
								x.push_back(_refPoint.x());
								x.push_back(_topRight.x());
								return x; }
	std::vector<T> pointsY() const { std::vector<T> y;
								y.push_back(_refPoint.y());
								y.push_back(_topRight.y());
								return y; }

	ShapeBasics::ShapeType type() const { return RECTANGLE; }
	std::string name() const { return "RECTANGLE"; }

private:
	Point<T> _topRight;
};


template<class T>
class RotRectangle : public virtual RotatedShape<T>,
					 public virtual Rectangle<T> {
public:
	RotRectangle() : Shape<T>(), RotatedShape<T>(), Rectangle<T>() {}
	RotRectangle(const Point<T> &refPoint, const Point<T> &topRight,
		   double rotAngle = 0., bool included = true) :
		   Shape<T>(refPoint, included),
		   RotatedShape<T>(refPoint, rotAngle, included),
		   Rectangle<T>(refPoint, topRight, included) {}
	~RotRectangle() {}

	ShapeBasics::ShapeType type() const { return ROTRECTANGLE; }
	std::string name() const { return "ROTRECTANGLE"; }
};


template<class T>
class Box : public virtual Shape<T> {
public:
  	Box() : Shape<T>(), _xWidth(0), _yWidth(0) {}
	//	referene point is center
  	Box(const Point<T> &refPoint, T xWidth, T yWidth,
  		 bool included = true) :
	 	Shape<T>(refPoint, included), _xWidth(xWidth), _yWidth(yWidth) {}
	~Box() {}

	T xWidth() const { return _xWidth; }
	T yWidth() const { return _yWidth; }
	std::vector<T> radii() const { std::vector<T> r;
							  r.push_back(_xWidth);
							  r.push_back(_yWidth);
							  return r; }

	ShapeBasics::ShapeType type() const { return BOX; }
	std::string name() const { return "BOX"; }

	inline bool contains(const Point<T> &p) const
	{
		T dx = ShapeUtils::ABS(p.x() - _refPoint.x()),
		  dy = ShapeUtils::ABS(p.y() - _refPoint.y());
		bool incTest = dx<=.5*_xWidth && dy<=.5*_yWidth;

		return included() ? incTest : !incTest;
	}

private:
	T _xWidth, _yWidth;
};

template<class T>
class RotBox : public virtual RotatedShape<T>,
			   public virtual Box<T> {
public:
	RotBox() : Shape<T>(), RotatedShape<T>(), Box<T>() {}
	RotBox(const Point<T> &refPoint, T xWidth, T yWidth,
		   double rotAngle = 0., bool included = true) :
		   Shape<T>(refPoint, included),
		   RotatedShape<T>(refPoint, rotAngle, included),
		   Box<T>(refPoint, xWidth, yWidth, included) {}
	~RotBox() {}

	ShapeBasics::ShapeType type() const { return ROTBOX; }
	std::string name() const { return "ROTBOX"; }

	std::string toSelectExpression(const std::string &xAxisName,
							  const std::string &yAxisName,
							  double binFactorX = 1.,
							  double binFactorY = 1.,
							  double minX = 1.,
							  double minY = 1.) const
			throw(UnexpressionableRegion);

};


template<class T>
class Rhombus : public virtual Box<T> {
public:
	Rhombus() : Shape<T>(), Box<T>() {}
	//	referene point is center
	Rhombus(const Point<T> &refPoint, T xWidth, T yWidth,
			  bool included = true) :
	  	Shape<T>(refPoint, included),
		Box<T>(refPoint, xWidth, yWidth, included) {}
	~Rhombus() {}

	ShapeBasics::ShapeType type() const { return RHOMBUS; }
	std::string name() const { return "RHOMBUS"; }
};


template<class T>
class RotRhombus : public virtual RotBox<T> {
public:
	RotRhombus() : Shape<T>(), RotatedShape<T>(), RotBox<T>() {}
	RotRhombus(const Point<T> &refPoint, T xWidth, T yWidth,
		   double rotAngle = 0., bool included = true) :
		   Shape<T>(refPoint, included),
		   RotatedShape<T>(refPoint, rotAngle, included),
		   Box<T>(refPoint, xWidth, yWidth, included),
		   RotBox<T>(refPoint, xWidth, yWidth, rotAngle, included) {}
	~RotRhombus() {}

	ShapeBasics::ShapeType type() const { return ROTRHOMBUS; }
	std::string name() const { return "ROTRHOMBUS"; }
};


template<class T>
class Sector : public virtual Shape<T> {
public:
	Sector() : Shape<T>(), _phiMin(0.), _phiMax(0.) {}
	Sector(const Point<T> &refPoint, double phiMin, double phiMax,
		   bool included = true) :
		   Shape<T>(refPoint, included),
		   _phiMin(phiMin), _phiMax(phiMax) {}

	double phiMin() const { return _phiMin; }
	double phiMax() const { return _phiMax; }
	std::vector<double> rotAngles() const { std::vector<double> a;
									   a.push_back(_phiMin);
									   a.push_back(_phiMax);
									   return a; }
	ShapeBasics::ShapeType type() const { return SECTOR; }
	std::string name() const { return "SECTOR"; }

private:
	double _phiMin, _phiMax;
};

template<class T>
class Polygon : public virtual Shape<T> {
public:
	Polygon() : Shape<T>(), _points(PointList<T>()) {}
	Polygon(const std::vector<T> &xCoords, const std::vector<T> &yCoords,
			unsigned nPoints, bool included = true) :
			// first point is reference point
			Shape<T>(Point<T>(xCoords[0], yCoords[0]), included),
			_points(xCoords, yCoords, nPoints)
	{
		//	check if the we have a terminal vertex, if yes, remove it
		if (nPoints<3)
			selerrstr<<"Cannot construct polygon from "<<nPoints
					 <<" point(s)"<<selerror("TooFewPoints", 1);
		for (unsigned n=1; n<nPoints; ++n)
			if (_points[n] == _points[0])
				//	found terminal vertex
				_points.resize(n);
	}
	~Polygon() {};

	ShapeBasics::ShapeType type() const { return POLYGON; }
	std::string name() const { return "POLYGON"; }

	const PointList<T> &vertices() const { return _points; }

	void resize(double factor);

	std::vector<T> pointsX() const
	{
		std::vector<T> v(_points.size());
		for (unsigned n=0; n<_points.size(); ++n)
			v[n] = _points[n].x();
		return v;
	}

	std::vector<T> pointsY() const
	{
		std::vector<T> v(_points.size());
		for (unsigned n=0; n<_points.size(); ++n)
			v[n] = _points[n].y();
		return v;
	}

	std::string toSelectExpression(const std::string &xAxisName,
								   const std::string &yAxisName,
								   double binFactorX = 1.,
								   double binFactorY = 1.,
								   double minX = 1.,
								   double minY = 1.) const
			throw(UnexpressionableRegion);

protected:
	PointList<T> _points;
};

typedef Point<float>			FPoint;
typedef Line<float>				FLine;
typedef Shape<float>			FShape;
typedef Circle<float>			FCircle;
typedef Sector<float>			FSector;
typedef Annulus<float>			FAnnulus;
typedef Ellipse<float>			FEllipse;
typedef ElliptAnnulus<float>	FElliptAnnulus;
typedef Box<float>				FBox;
typedef RotBox<float>			FRotBox;
typedef Rectangle<float>		FRectangle;
typedef RotRectangle<float>		FRotRectangle;
typedef Rhombus<float>			FRhombus;
typedef RotRhombus<float>		FRotRhombus;
typedef Polygon<float>			FPolygon;

//
//	A Component a geometrical area formed by the union of a list of a list of
//	constituent shapes, i.e. a point lies within the Component if it lies in
//	at least one of its constituent	shapes (logical OR)
//
class Component : public FShape, public std::vector<FShape *> {
public:
	Component() {}
	Component(unsigned n) : std::vector<FShape *>(n) {}
	Component(unsigned n, FShape *val) : std::vector<FShape *>(n, val) {}

	std::string toSelectExpression(const std::string &xAxisName,
								   const std::string &yAxisName,
								   double binFactorX = 1.,
								   double binFactorY = 1.,
								   double minX = 1.,
								   double minY = 1.) const
												throw(UnexpressionableRegion);

	bool contains(const FPoint &) const;
};

//
//	A GeoRegion is a geometrical area formed by the intersection of a list
//	of constituent Components (see above), i.e. a point lies within the
//	GeoRegion if it lies in all of its constituent Components (logical AND)
//
class GeoRegion : public FShape, public std::vector<Component> {
public:
	class InvalidRegion {};	//	exception class to be thrown
								//	by objects which construct a region
								//	object and find an error

	//	default constructor - create empty list
	GeoRegion() : FShape() {}

	//	destructor
	~GeoRegion();

	void free();
		//	free up resources allocated by object - the object is no
		//	longer usable

	unsigned elements() const;
	bool hasPolygon() const; //	has the region at least one polygon?
	unsigned maxPoints() const;
	unsigned maxRadii() const;
	unsigned maxRotAngles() const;

	std::string toSelectExpression(const std::string &xAxisName,
								   const std::string &yAxisName,
								   double binFactorX = 1.,
								   double binFactorY = 1.,
								   double minX = 1.,
								   double minY = 1.) const
												throw(UnexpressionableRegion);

	bool contains(const FPoint &) const;
};


};

using namespace SelectlibRegion;

#endif
