//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	TriangulatablePolygon.h: --- A polygon that can be triangulated
//
//	DESCRIPTION:
//		A TriangulatablePolygon is a polygon that can be decomposed into a
//		finite list of triangles. This list can be used for a variety of
//		geometrical operations with this polygon but can also serve as
//		the basis for a fast O(log N) point-in/out test. N is the number
//		of polygon vertices. A triangulation of a N-vertex polygon
//		gives a triangle list of size N-2.
//
//	CHANGE HISTORY:
//		UL, 2001-08-01: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Wed Aug  1 11:40:15 MET DST 2001
//
//	REFERENCES:
//		[1] http://www.cs.unc.edu/~dm/CODE/GEM/chapter.html
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/TriangulatablePolygon.h,v 1.4 2002/12/17 17:48:12 jbakker Exp $
//

#ifndef TRIANGULATABLE_POLYGON_H
#define TRIANGULATABLE_POLYGON_H

#include "Region.h"

class TriangulatablePolygon : public Polygon<double> {
public:
	class IntersectingSegments {};
	TriangulatablePolygon(const std::vector<double> &xCoords,
						  const std::vector<double> &yCoords,
						  bool pointsScattered = false,
						  bool testSegments = false)
		throw(IntersectingSegments);
		//
		//	PURPOSE:
		//		Construct and triangulate polyon from given list
		//		of vectices
		//
		//	ARGUMENTS:
		//		xCoords, yCoords	: vector with X/Y coordinates of vertices
		//		pointsScattered		: true: no particular order of the
		//								    vetrices is assumed; will establish
		//									an order by constructing a simple
		//									closed path
		//							  false: the vertex order is supposed
		//									 to be clock- or anticlockwise
		//		testSegments		: an intrinsic property of the
		//							  point-in-polygon test is that the result
		//							  for test points that lie exactly on a
		//							  segment or coincide with a vertex is
		//							  uncertain. Setting testSegments=true
		//							  ensures border-points will be correctly
		//							  treated as belonging to the polygon.
		//							  Please note: This comes at the price
		//							  a degraded run time efficiency, so,
		//							  if the intrinsic behavior is acceptable
		//							  it is advised to leave testSegments=false
		//
		//	EXCEPTIONS:
		//		IntersectingSegments: thrown if pointsScattered=false
		//		and intersecting segments are found; in this case no
		//		triangulatable polygon can be construced
		//


	bool contains(const Point<double> &) const;
		//
		//	PURPOSE:
		//		Point-in-polygon test; please note that the result for
		//		points exactly on the boundary is not consistent! This
		//		is a limitation that is hard to avoid - it certainly
		//		can but the benefit does not outweigh the loss of
		//		efficiency.
		//


	typedef int (*TriangleList)[3];
	unsigned triangles() const;	//	how many triangles does the list
								//	consist of?
	const TriangleList &triangulation() const;
		//
		//	PURPOSE:
		//		Return the result of the triangulation as an array
		//		of three-tuple triangle indices. The size of the array
		//		can be inquired with triangles()
		//


	~TriangulatablePolygon();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	mutable bool _triangulated;			//	Has this been triangulated yet?
	mutable TriangleList _triangleList;	//	Fixed triangle list
	mutable unsigned _triangles;		//	length of _triangleList
	mutable bool _testSegments;			//	test segments
	std::vector<Line<double> > _segments;	//	segments

public:
	typedef struct {
		double x, y;
	} point_t, vector_t;

private:
	//	Segment attributes
	typedef struct {	
		point_t	v0, v1;			// two endpoints
		bool	is_inserted;	// inserted in trapezoidation yet ?
		int		root0, root1;	// root nodes in Q
		int		next;			// Next logical segment
		int		prev;			// Previous segment
	} segment_t;

	//	Trapezoid attributes
	typedef struct {
		int lseg, rseg;			// two adjoining segments
		point_t hi, lo;			// max/min y-values
		int u0, u1;
		int d0, d1;
		int sink;				// pointer to corresponding in Q
		int usave, uside;		// I forgot what this means
		int state;
	} trap_t;

	//	Node attributes for every node in the query structure
	typedef struct {
		int nodetype;			// Y-node or S-node
		int segnum;
		point_t yval;
		int trnum;
		int parent;				// doubly linked DAG
		int left, right;		// children
	} node_t;

	//	Circularly linked list describing the monotone polygon
	typedef struct {
		int vnum;
		int next;
		int prev;
		bool marked;
	} monchain_t;

	typedef struct {
		point_t pt;
		int vnext[4];			// next vertices for the 4 chains
		int vpos[4];			// position of v in the 4 chains
		int nextfree;
	} vertexchain_t;

	//	data
	node_t		*_qs;			// Query structure
	trap_t		*_tr;			// Trapezoid structure
	segment_t	*_seg;			// Segment table
	mutable int	*_permute;		// permutation indices

	mutable monchain_t	*_mchain;
								// Table to hold all the monotone
								// polygons . Each monotone polygon
								// is a circularly linked list

	mutable vertexchain_t *_vert;
								// chain init. information. This
								// is used to decide which
								// monotone polygon to split if
								// there are several other
								// polygons touching at the same
								// vertex

	mutable int			*_mon;	// contains position of any vertex in
								// the monotone chain for the polygon

	mutable bool		*_visited;

	mutable int			_mon_idx;
	mutable int			_chain_idx;
	mutable int			_op_idx;

	mutable int			_choose_idx,
						_q_idx,
						_tr_idx;

	//	array sizes
	unsigned SEGSIZE,	//	max# of segments. Determines how
						//	many points can be specified as
						//	input. If your datasets have large
						//	number of points, increase this
						//	value accordingly
			 QSIZE,		//	maximum table sizes
			 TRSIZE;	//	max# trapezoids

	//	methods
	void initialise(unsigned) const;
	void generate_random_ordering(unsigned) const;
	int choose_segment() const;
	void construct_trapezoids(int) const;
	int init_query_structure(int) const;
	void add_segment(int) const;
	void merge_trapezoids(int, int, int, int) const;
	void find_new_roots(int) const;
	int newnode() const;
	int newtrap() const;
	int newmon() const;
	int new_chain_element() const;
	bool inserted(int, int) const;
	unsigned monotonate_trapezoids(unsigned) const;
	int triangulate_monotone_polygons(unsigned, unsigned, TriangleList) const;
	int triangulate_single_polygon(unsigned, unsigned, unsigned,
									TriangleList) const;
	int traverse_polygon(int, int, int, int);
	int locate_endpoint(point_t *, point_t *, int) const;
	void get_vertex_positions(int, int, int *, int *) const;
	int make_new_monotone_poly(int, int, int) const;
	int traverse_polygon(int, int, int, int) const;
	bool is_left_of(int, point_t *) const;
	bool inside_polygon(trap_t *) const;

	bool liesOnSegment(const Point<double> &p) const;
};

#endif
